# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import random
from itertools import islice
from typing import Dict, List, Optional, Tuple

import requests
import yaml
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from tqdm import tqdm

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def save_yaml(data: Dict, file_path: str):
    with open(file_path, "w") as file:
        yaml.dump(data, file, default_flow_style=False)


def split_train_val_sequences_with_ratio(
    sequences_tar_info_dict: Dict[str, Dict[str, List[str]]], ratio: float
) -> Tuple[Dict[str, Dict[str, List[str]]], Dict[str, Dict[str, List[str]]]]:
    """
    Given a dictionary of sequences, split it into train and validation sets based on the specified ratio and random seed.
    """
    sequences = list(sequences_tar_info_dict.keys())
    split_point = int(len(sequences) * ratio)
    train_sequences = sequences[:split_point]
    valid_sequences = sequences[split_point:]
    train_sequences_tar_info_dict = {
        sequence: sequences_tar_info_dict[sequence] for sequence in train_sequences
    }
    validation_sequences_tar_info_dict = {
        sequence: sequences_tar_info_dict[sequence] for sequence in valid_sequences
    }
    return train_sequences_tar_info_dict, validation_sequences_tar_info_dict


def split_train_val_sequences_by_json(
    sequences_tar_info_dict: Dict, train_val_split_json_path: str
) -> Tuple[Dict, Dict]:
    """
    Given a dictionary of sequences, split it into train and validation sets based on the specified json file.
    """
    with open(train_val_split_json_path, "r") as file:
        train_val_split_data = json.load(file)
        train_tars_sequence_names = train_val_split_data["train_sequences"]
        validation_tars_sequence_names = train_val_split_data["val_sequences"]
        train_sequences = find_tar_info_with_sequence_names(
            sequences_tar_info_dict, train_tars_sequence_names
        )
        validation_sequences = find_tar_info_with_sequence_names(
            sequences_tar_info_dict, validation_tars_sequence_names
        )
    return train_sequences, validation_sequences


def extract_first_K_sequences(
    wds_file_urls: Dict, max_num_sequences: Optional[int] = None
) -> Dict:
    """
    Extract tar info from the given dictionary of WDS file info.
    If max_num_sequences is specified, only the first max_num_sequences sequences will be included.
    """
    # if no cropping is needed
    if not max_num_sequences or max_num_sequences < 0:
        return wds_file_urls
    else:
        return dict(islice(wds_file_urls.items(), max_num_sequences))


def download_wds_files_for_single_sequence(
    urls: Dict[str, Dict[str, str]], output_dir: str, sequence_name: str
) -> Tuple[List[str], List[str]]:
    """
    For each sequence, download all the files to a directory with the same name.
    Return a dictionary mapping sequence names to lists of downloaded file paths.
    """
    successful_urls = []
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    # Set up retry strategy
    retries = Retry(
        total=5,
        backoff_factor=2,  # final try is 2*(2**5) = 64 seconds
        status_forcelist=[429, 500, 502, 503, 504],
    )
    adapter = HTTPAdapter(max_retries=retries)
    http = requests.Session()
    http.mount("http://", adapter)
    http.mount("https://", adapter)
    failed_urls = []
    for tar_name, tar_info in urls.items():
        # original tar name is shards-0000_tar, we need to change to shards-0000.tar
        tar_name = tar_name.replace("_tar", ".tar")
        filepath = os.path.join(output_dir, tar_name)
        relative_path = os.path.join(sequence_name, tar_name)
        # Initialize the list for this sequence if not already done
        try:
            response = http.get(tar_info["download_url"], stream=True)
            response.raise_for_status()  # Raise an exception for bad status codes
            directory = os.path.dirname(filepath)
            if not os.path.exists(directory):
                os.makedirs(directory)
            with open(filepath, "wb") as f:
                for chunk in response.iter_content(1024):
                    f.write(chunk)
            successful_urls.append(relative_path)
        except requests.exceptions.RequestException as e:
            logger.error(f"Failed to download {relative_path}: {e}")
            failed_urls.append(relative_path)
    return successful_urls, failed_urls


def download_sequences(
    download_sequences_tar_info_dict: Dict[str, Dict[str, List[str]]],
    output_folder_path: str,
) -> Tuple[Dict, Dict]:
    """
    Each sequence will be downloaded into a new directory named after the sequence within the specified output folder.
    The function logs the download progress, creates necessary directories, and uses the `download_wds_files_for_single_sequence` function
    to actually download the files.
    """
    success_sequences_with_tars = {}
    failed_sequences_with_tars = {}
    # Wrap the loop with tqdm for a progress bar
    for sequence_name, sequence_info in tqdm(
        download_sequences_tar_info_dict.items(), desc="Downloading sequences"
    ):
        logger.info(f"Downloading {sequence_name}")
        sequence_dir = os.path.join(output_folder_path, sequence_name)
        os.makedirs(
            sequence_dir, exist_ok=True
        )  # Added exist_ok=True to avoid error if directory already exists
        successful_tar_list, failed_tar_list = download_wds_files_for_single_sequence(
            sequence_info, sequence_dir, sequence_name
        )
        success_sequences_with_tars[sequence_name] = successful_tar_list
        if len(failed_tar_list) > 0:
            failed_sequences_with_tars[sequence_name] = failed_tar_list
    return failed_sequences_with_tars, success_sequences_with_tars


def get_url_from_tars_info(tars_info: Dict) -> Dict:
    # extract tar urls from tars_info, which is a dictionary
    tar_urls = {}
    for sequence_name, sequence_tars_info in tars_info.items():
        tar_urls[sequence_name] = [
            tar_info["download_url"] for tar_info in sequence_tars_info.values()
        ]
    return tar_urls


def write_tar_yaml(
    output_folder_path,
    train_tars,
    validation_tars,
    all_tars,
    train_tars_yaml_name,
    validation_tars_yaml_name,
    all_tars_yaml_name,
):
    with open(os.path.join(output_folder_path, train_tars_yaml_name), "w") as file:
        yaml.dump({"tars": train_tars}, file, default_flow_style=False)

    with open(os.path.join(output_folder_path, validation_tars_yaml_name), "w") as file:
        yaml.dump({"tars": validation_tars}, file, default_flow_style=False)
    with open(os.path.join(output_folder_path, all_tars_yaml_name), "w") as file:
        yaml.dump({"tars": all_tars}, file, default_flow_style=False)
    logger.info(
        f"{train_tars_yaml_name}, {validation_tars_yaml_name}, {all_tars_yaml_name} are written successfully to {output_folder_path}."
    )


def find_tar_info_with_sequence_names(
    tars_info: Dict, sequence_names: List[str]
) -> Dict:
    """
    Find the tar info with the specified sequence names in the given dictionary.
    """
    result = {}
    for sequence_name in sequence_names:
        if sequence_name in tars_info.keys():
            result[sequence_name] = tars_info[sequence_name]
    return result


def split_train_val_sequences(
    sequences_tar_info_dict: Dict,
    train_val_split_json_path: Optional[str],
    train_val_split_ratio: Optional[float],
) -> Tuple[Dict, Dict]:
    """
    split data, user has two options:
    1. use the train_val_split_json_path to specify the train and validation split
    2. use the train_val_split_ratio to split the data
    """
    if train_val_split_json_path is not None:
        train_sequences, validation_sequences = split_train_val_sequences_by_json(
            sequences_tar_info_dict, train_val_split_json_path
        )
    else:
        assert train_val_split_ratio is not None, "train_val_split_ratio is None"
        train_sequences, validation_sequences = split_train_val_sequences_with_ratio(
            sequences_tar_info_dict, train_val_split_ratio
        )
    return train_sequences, validation_sequences


def randomize_sequence_order(sequences_tar_info_dict: Dict, random_seed: int) -> Dict:
    """
    Randomize the order of sequences in the given dictionary.
    """
    random.seed(random_seed)
    sequences = list(sequences_tar_info_dict.keys())
    sequences.sort()
    random.shuffle(sequences)
    randomized_sequences_tar_info_dict = {
        sequence: sequences_tar_info_dict[sequence] for sequence in sequences
    }
    return randomized_sequences_tar_info_dict


def download_atek_wds_sequences(
    config_name: str,
    input_json_path: str,
    train_val_split_ratio: float,
    random_seed: int,
    output_folder_path: str,
    max_num_sequences: Optional[int] = None,
    download_wds_to_local: bool = False,
    train_val_split_json_path: Optional[str] = None,
):
    # extract tar urls
    with open(input_json_path, "r") as file:
        data = json.load(file)
    sequence_to_tar_info = data["atek_data_for_all_configs"][config_name][
        "wds_file_urls"
    ]

    sequence_to_tar_info = randomize_sequence_order(sequence_to_tar_info, random_seed)

    sequences_tar_info_dict = extract_first_K_sequences(
        sequence_to_tar_info, max_num_sequences
    )

    train_sequences_tar_info_dict, validation_sequences_tar_info_dict = (
        split_train_val_sequences(
            sequences_tar_info_dict,
            train_val_split_json_path,
            train_val_split_ratio,
        )
    )
    # Save the tars failed to download
    if not os.path.exists(output_folder_path):
        os.makedirs(output_folder_path)
    if download_wds_to_local is True:
        # Download sequences to local
        failed_train_sequence_with_tars, successful_train_sequences_with_tars = (
            download_sequences(train_sequences_tar_info_dict, output_folder_path)
        )
        (
            failed_validation_sequence_with_tars,
            successful_validation_sequences_with_tars,
        ) = download_sequences(validation_sequences_tar_info_dict, output_folder_path)

        successful_all_sequences_with_tars = (
            successful_train_sequences_with_tars
            | successful_validation_sequences_with_tars
        )
        failed_sequences_with_tars = (
            failed_train_sequence_with_tars | failed_validation_sequence_with_tars
        )

        # Save the YAML files
        with open(
            os.path.join(output_folder_path, "local_train_tars.yaml"), "w"
        ) as file:
            yaml.dump(
                {"tars": successful_train_sequences_with_tars},
                file,
                default_flow_style=False,
            )
        with open(
            os.path.join(output_folder_path, "local_validation_tars.yaml"), "w"
        ) as file:
            yaml.dump(
                {"tars": successful_validation_sequences_with_tars},
                file,
                default_flow_style=False,
            )
        with open(os.path.join(output_folder_path, "local_all_tars.yaml"), "w") as file:
            yaml.dump(
                {"tars": successful_all_sequences_with_tars},
                file,
                default_flow_style=False,
            )
        logger.info(
            f"Successfully download {len(successful_all_sequences_with_tars)} sequences."
        )

        # If failed to download any sequences, log the results
        if len(failed_sequences_with_tars) > 0:
            logger.warning(
                f"Failed to download {len(failed_sequences_with_tars)} sequences."
            )
            for sequence_name, failed_tars in failed_sequences_with_tars.items():
                logger.info(
                    f"Failed to download {sequence_name}: {len(failed_tars)} tars."
                )

    else:
        # original train_tars, valid_tars and tar_urls save the file name, sha1sum, file size and download url, now we only need the download url
        train_sequences_tar_url = get_url_from_tars_info(train_sequences_tar_info_dict)
        validation_sequences_tar_url = get_url_from_tars_info(
            validation_sequences_tar_info_dict
        )
        all_sequences_tar_url = train_sequences_tar_url | validation_sequences_tar_url
        write_tar_yaml(
            output_folder_path=output_folder_path,
            train_tars=train_sequences_tar_url,
            validation_tars=validation_sequences_tar_url,
            all_tars=all_sequences_tar_url,
            train_tars_yaml_name="streamable_train_tars.yaml",
            validation_tars_yaml_name="streamable_validation_tars.yaml",
            all_tars_yaml_name="streamable_all_tars.yaml",
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import json
import os
from typing import Dict, List

import numpy as np
import pandas as pd
from projectaria_tools.core.sophus import SE3


class AdtPredictionSaver:
    """
    Save 3D bounding box model predictions to files of the ADT format.

    Args:
        output_dir (str): directory where the predictions will be saved
        category_id_remapping_json (str): path to the json file containing the mapping from
            instance ID to category ID
    """

    def __init__(
        self,
        output_dir: str,
        metadata_file: str,
        category_id_remapping_json: str,
    ):
        self.output_dir = output_dir

        with open(metadata_file, "r") as f:
            self.instance_metadata = json.load(f)

        with open(category_id_remapping_json, "r") as f:
            inst_ids_to_cat_ids = json.load(f)
            self.cat_ids_to_inst_ids = {
                int(cat_id): int(inst_id)
                for inst_id, cat_id in inst_ids_to_cat_ids.items()
            }

    def __call__(self, prediction_list: List[List[List[Dict]]]):
        # convert prediction to pandas dataframe for processing
        flattend_prediction_list = itertools.chain(*prediction_list)
        predictions_df = pd.DataFrame(itertools.chain(*flattend_prediction_list))

        # create directory to save the predictions
        seq_name = predictions_df.iloc[0]["sequence_name"]
        pred_dir = os.path.join(self.output_dir, seq_name)
        if not os.path.isdir(pred_dir):
            os.makedirs(pred_dir)

        cat_ids = predictions_df.loc[:, "category_idx"].tolist()
        object_uids = [self.cat_ids_to_inst_ids[cat_id] for cat_id in cat_ids]

        self._save_3d_bbox(predictions_df, object_uids, pred_dir)
        self._save_2d_bbox(predictions_df, object_uids, pred_dir)
        self._save_scene_objects(predictions_df, object_uids, pred_dir)
        self._save_instances(object_uids, pred_dir)

    def _save_3d_bbox(
        self, predictions_df: pd.DataFrame, object_uids: List[int], pred_dir: str
    ):
        dimensions = np.stack(predictions_df.loc[:, "dimensions"])
        timestamps = predictions_df.loc[:, "timestamp_ns"].tolist()
        dimension_data = {
            "object_uid": object_uids,
            "timestamp[ns]": timestamps,
            "p_local_obj_xmin[m]": -dimensions[:, 0] / 2,
            "p_local_obj_xmax[m]": dimensions[:, 0] / 2,
            "p_local_obj_ymin[m]": -dimensions[:, 1] / 2,
            "p_local_obj_ymax[m]": dimensions[:, 1] / 2,
            "p_local_obj_zmin[m]": -dimensions[:, 2] / 2,
            "p_local_obj_zmax[m]": dimensions[:, 2] / 2,
        }
        dimension_data = pd.DataFrame(dimension_data)
        dimension_data.to_csv(f"{pred_dir}/3d_bounding_box.csv", index=False)

    def _save_2d_bbox(
        self, predictions_df: pd.DataFrame, object_uids: List[int], pred_dir: str
    ):
        timestamps = predictions_df.loc[:, "timestamp_ns"].tolist()
        bbox_2D = np.stack(predictions_df.loc[:, "bbox_2D"])
        bbox_2d_data = {
            "stream_id": ["214-1"] * len(object_uids),
            "object_uid": object_uids,
            "timestamp[ns]": timestamps,
            "x_min[pixel]": bbox_2D[:, 0],
            "x_max[pixel]": bbox_2D[:, 2],
            "y_min[pixel]": bbox_2D[:, 1],
            "y_max[pixel]": bbox_2D[:, 3],
        }
        bbox_2d_data = pd.DataFrame(bbox_2d_data)
        bbox_2d_data.to_csv(f"{pred_dir}/2d_bounding_box.csv", index=False)

    def _save_scene_objects(
        self, predictions_df: pd.DataFrame, object_uids: List[int], pred_dir: str
    ):
        # compute object pose in world
        timestamps = predictions_df.loc[:, "timestamp_ns"].tolist()
        ts_cam_obj = np.stack(predictions_df.loc[:, "t_cam_obj"])
        Rs_cam_obj = np.stack(predictions_df.loc[:, "R_cam_obj"])
        Ts_world_cam = predictions_df.loc[:, "T_world_cam"]
        confidence = np.stack(predictions_df.loc[:, "score"])
        translations = []
        quaternions = []
        for R_cam_obj, t_cam_obj, T_world_cam in zip(
            Rs_cam_obj, ts_cam_obj, Ts_world_cam
        ):
            T_cam_obj = np.concatenate((R_cam_obj, t_cam_obj[:, np.newaxis]), axis=1)
            T_world_obj = SE3.from_matrix3x4(T_world_cam) @ SE3.from_matrix3x4(
                T_cam_obj
            )

            trans = T_world_obj.translation().squeeze()
            quat = T_world_obj.rotation().to_quat().squeeze()
            translations.append(trans)
            quaternions.append([quat[0], quat[1], quat[2], quat[3]])
        translations_np = np.array(translations)
        quaternions_np = np.array(quaternions)

        scene_objects_data = {
            "object_uid": object_uids,
            "timestamp[ns]": timestamps,
            "t_wo_x[m]": translations_np[:, 0],
            "t_wo_y[m]": translations_np[:, 1],
            "t_wo_z[m]": translations_np[:, 2],
            "q_wo_w": quaternions_np[:, 0],
            "q_wo_x": quaternions_np[:, 1],
            "q_wo_y": quaternions_np[:, 2],
            "q_wo_z": quaternions_np[:, 3],
        }
        scene_objects_data["confidence"] = confidence
        scene_objects_data = pd.DataFrame(scene_objects_data)
        scene_objects_data.to_csv(f"{pred_dir}/scene_objects.csv", index=False)

    def _save_instances(self, object_uids: List[int], pred_dir: str):
        unique_object_uids = set(object_uids)
        instances_data = {
            str(uid): self.instance_metadata[str(uid)] for uid in unique_object_uids
        }
        with open(f"{pred_dir}/instances.json", "w") as f:
            json.dump(instances_data, f, indent=2)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argparse import Namespace
from typing import Callable, Dict, List, Union

from atek_v1.evaluation.bbox3d_evaluator import Bbox3DEvaluator
from atek_v1.inference.adt_prediction_saver import AdtPredictionSaver
from atek_v1.inference.cubercnn_postprocessor import CubercnnPredictionConverter
from atek_v1.viz.visualization_callbacks import AtekInferViewer


def create_inference_callback(
    args: Namespace, model_config: Dict
) -> Dict[str, Union[Callable, List[Callable]]]:
    """
    Create callback functions for the inference pipeline, such as visualization and saving model
    predictions, based on model arhitecture and dataset requirements.

    Args:
        args (Namespace): args with options to create callbacks, such as args.visualize
        model_config (Dict): model configs with options to creating the callbacks

    Returns:
        callbacks (Dict[str, Union[Callable, List[Callable]]): a dict of callback functions.
            `post_processor` converts model output to another format. `per_batch_callbacks` are
            called after each iteration, while `per_sequence_callbacks` are called at the end of each
            webdataset tar or raw video sequence.
    """

    if args.model_name == "cubercnn":
        post_processor = CubercnnPredictionConverter(
            model_config["score_threshold"], model_config["category_names"]
        )

        per_batch_callbacks = []
        if args.visualize:
            per_batch_callbacks.append(
                AtekInferViewer(args.web_port, args.ws_port, "camera_rgb")
            )

        per_sequence_callbacks = []
        if args.save_prediction:
            per_sequence_callbacks.append(
                AdtPredictionSaver(
                    args.output_dir,
                    args.metadata_file,
                    model_config["cfg"].ID_MAP_JSON,
                )
            )

        eval_callbacks = []
        if args.evaluate:
            eval_callback = Bbox3DEvaluator(args.metric_name, args.metric_threshold)
            eval_callbacks.append(eval_callback)

        callbacks = {
            "post_processor": post_processor,
            "per_batch_callbacks": per_batch_callbacks,
            "per_sequence_callbacks": per_sequence_callbacks,
            "eval_callbacks": eval_callbacks,
        }
    else:
        raise ValueError(
            f"Unknown model architecture for setting callbacks: {args.model_name}"
        )

    return callbacks



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-strict

import logging
import os
from typing import Dict, List, Optional, Tuple

import numpy as np

import torch
from atek.evaluation.static_object_detection.eval_obb3_metrics_utils import (
    prec_recall_bb3,
)
from atek.evaluation.static_object_detection.obb3_csv_io import AtekObb3CsvReader
from atek.evaluation.static_object_detection.static_object_detection_metrics import (
    AtekObb3Metrics,
)
from atek.util.atek_constants import ATEK_CATEGORY_ID_TO_NAME
from atek.util.tensor_utils import filter_obbs_by_confidence

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# TODO: make this a input from file
sem_id_to_name = ATEK_CATEGORY_ID_TO_NAME


def compute_prec_recall_for_single_timestamp(
    pred_obb_dict: Dict,
    gt_obb_dict: Dict,
    iou: float = 0.2,
    compute_per_class_metrics: bool = False,
) -> Dict:
    """
    Compute precision and recall for a single timestamp, given the prediction and gt obbs.
    """
    result = {}

    # perform precision recall calculation for single timestamp
    tupled_results = prec_recall_bb3(
        pred_obb_dict,
        gt_obb_dict,
        iou_thres=iou,
        return_ious=True,
        per_class=compute_per_class_metrics,
    )
    precision, recall, match_matrix, *extra_results = tupled_results
    per_class_results = extra_results[1] if len(extra_results) > 1 else None

    # Compute stats including true positives, false positives from precision and recall
    true_positives = match_matrix.any(-1)

    # Log results into a dict
    result[f"precision@IoU{iou}"] = float(precision)
    result[f"recall@IoU{iou}"] = float(recall)
    result[f"num_true_positives@IoU{iou}"] = int(true_positives.sum())
    result["num_predictions"] = match_matrix.shape[0]
    result["num_groundtruth"] = match_matrix.shape[1]

    # Log per-class results
    if compute_per_class_metrics:
        for sem_id, per_class_result in per_class_results.items():
            result[f"precision@IoU{iou}@Class_{sem_id_to_name[sem_id.item()]}"] = float(
                per_class_result["precision"]
            )
            result[f"recall@IoU{iou}@Class_{sem_id_to_name[sem_id.item()]}"] = float(
                per_class_result["recall"]
            )

    return result


def update_from_single_sequence_obb3(
    pred_obb_dict: Dict,
    gt_obb_dict: Dict,
    mAP_3d: AtekObb3Metrics,
    iou: float,
    single_timestamp_to_log_result: Optional[int] = None,
    compute_per_class_metrics: bool = False,
    confidence_lower_threshold: float = 0.3,
) -> Tuple[Dict, List, List]:
    """
    Core function to evaluate a single pair of obb gt vs prediction, for a single sequence,
    and update the mAP_3d metrics accordingly.

    @param pred_obb_dict: a dict of {timestamp -> Obb3GtDict} for the prediction obbs, need to include "confidence_scores"
    @param gt_obb_dict: a dict of {timestamp -> Obb3GtDict} for the gt
    @param mAP_3d: the mAP_3d metrics to be updated.
    @param single_timestamp_to_log_result: if not None, log the precision and recall results for the single frame specified by this timestamp.
    """
    result = {}

    # Aggregate all timestamps
    all_timestamps = list(pred_obb_dict.keys()) + list(gt_obb_dict.keys())
    all_timestamps = list(set(all_timestamps))
    all_timestamps.sort()

    # Count for timestamps with empty GT and empty predictions
    empty_gt_timestamp_count = 0
    empty_pred_timestamp_count = 0
    all_precisions = []
    all_recalls = []
    for time in all_timestamps:
        if time not in pred_obb_dict:
            logger.info(f"prediction obbs not found for timestamp {time}")
            empty_pred_timestamp_count += 1
            continue
        if time not in gt_obb_dict:
            logger.info(f"gt obbs not found for {time}")
            empty_gt_timestamp_count += 1
            continue

        # check if the preds contain confidence scores
        confidence_score = pred_obb_dict[time]["confidence_scores"].squeeze()
        assert not torch.all(
            confidence_score.eq(-1.0)
        ), "the obbs don't contain valid confidence scores for mAP calculation."

        # TODO: check paddings are removed for EFM data

        # filter prediction's confidence scores with a lower threshold
        filtered_pred_obb_dict = filter_obbs_by_confidence(
            pred_obb_dict[time],
            confidence_score=confidence_score,
            confidence_lower_threshold=confidence_lower_threshold,
        )
        # Skip if no prediction obbs left after filtering
        if (
            not filtered_pred_obb_dict
            or filtered_pred_obb_dict["category_ids"].numel() == 0
        ):
            logger.debug(
                f"no prediction obbs left after filtering for timestamp {time}, marking as empty pred timestamp"
            )
            empty_pred_timestamp_count += 1
            continue

        single_timestamp_prec_recall_result = compute_prec_recall_for_single_timestamp(
            filtered_pred_obb_dict,
            gt_obb_dict[time],
            iou=iou,
            compute_per_class_metrics=compute_per_class_metrics,
        )
        all_precisions.append(
            single_timestamp_prec_recall_result[f"precision@IoU{iou}"]
        )
        all_recalls.append(single_timestamp_prec_recall_result[f"recall@IoU{iou}"])

        # log precision-recall statistics for a single frame if needed
        if (
            single_timestamp_to_log_result is not None
            and time == single_timestamp_to_log_result
        ):
            logger.info(f"logging eval result for a single frame {time} ... ")
            single_timestamp_prec_recall_result = (
                compute_prec_recall_for_single_timestamp(
                    pred_obb_dict[time],
                    gt_obb_dict[time],
                    iou=iou,
                    compute_per_class_metrics=compute_per_class_metrics,
                )
            )
            result.update(single_timestamp_prec_recall_result)

        # add pred/gt pair to mAP calculator.
        mAP_3d.update(pred_obb_dict[time], gt_obb_dict[time])

    # End for loop over timestamps
    result["num_timestamps"] = len(all_timestamps)
    result["num_timestamp_miss_pred"] = empty_pred_timestamp_count
    result["num_timestamp_miss_gt"] = empty_gt_timestamp_count
    result[f"precision@IoU{iou},Confidence{confidence_lower_threshold}"] = np.mean(
        all_precisions
    )
    result[f"recall@IoU{iou},Cofidence{confidence_lower_threshold}"] = np.mean(
        all_recalls
    )

    return result, all_precisions, all_recalls


def evaluate_obb3_for_single_csv_pair(
    pred_csv: str,
    gt_csv: str,
    iou: float = 0.2,
    log_last_frame_result: bool = False,
    compute_per_class_metrics: bool = False,
    confidence_lower_threshold: float = 0.3,
) -> Dict:
    """
    Evaluate a single pair of prediction and gt obbs csv files, and return a dict of metrics including mean average precision (mAP), precision, recall based on IOU value.
    """
    # Read in prediction and gt obbs
    pred_reader = AtekObb3CsvReader(input_filename=pred_csv)
    pred_obb_dict = pred_reader.read_as_obb_dict()
    gt_reader = AtekObb3CsvReader(input_filename=gt_csv)
    gt_obb_dict = gt_reader.read_as_obb_dict()

    # TODO: check or allow different taxonomies from prediction and gt

    result = {}
    mAP_3d = AtekObb3Metrics(
        class_metrics=compute_per_class_metrics,
        global_name_to_id={
            name: int(sem_id) for sem_id, name in sem_id_to_name.items()
        },
    )

    # If last frame result needs to be logged
    timestamp_to_log = None
    if log_last_frame_result:
        timestamp_to_log = max(pred_obb_dict.keys()) if pred_obb_dict else None

    single_sequence_result, _, _ = update_from_single_sequence_obb3(
        pred_obb_dict=pred_obb_dict,
        gt_obb_dict=gt_obb_dict,
        mAP_3d=mAP_3d,
        iou=iou,
        compute_per_class_metrics=compute_per_class_metrics,
        confidence_lower_threshold=confidence_lower_threshold,
        single_timestamp_to_log_result=timestamp_to_log,
    )
    result.update(single_sequence_result)

    # Compute mAP metrics numbers, ignore average recall (e.g. "mar_*")
    result_map = mAP_3d.compute()
    result_map = {
        k: v.item() for k, v in result_map.items() if not k.startswith("mar_")
    }
    result.update(result_map)

    return result


def evaluate_obb3_over_a_dataset(
    input_folder: str,
    gt_filename: str,
    prediction_filename: str,
    iou: float = 0.2,
    compute_per_class_metrics: bool = False,
    confidence_lower_threshold: float = 0.3,
    max_num_sequences: int = -1,
) -> Dict:
    """
    Perform static object detection evaluation at a dataset level.
    The input_folder is a folder containing a list of sequences, each of which contains a pair of prediction and gt obb csv files.
    """
    # get all the pred and gt csv files
    pred_csv_paths, gt_csv_paths = [], []
    sequence_names = os.listdir(input_folder)
    dirs = [os.path.join(input_folder, seq) for seq in sequence_names]
    dirs = [d for d in dirs if os.path.isdir(d)]
    dirs = sorted(dirs)

    # Limit number of sequences
    if max_num_sequences > 0:
        dirs = dirs[0:max_num_sequences]

    for d in dirs:
        pred_csv = os.path.join(d, prediction_filename)
        gt_csv = os.path.join(d, gt_filename)
        if os.path.exists(gt_csv) and os.path.exists(pred_csv):
            pred_csv_paths.append(pred_csv)
            gt_csv_paths.append(gt_csv)

    # If no gt or pred csv files are found, return an almost-empty dict
    result = {}
    result["num_seqs"] = len(pred_csv_paths)
    if len(pred_csv_paths) == 0 or len(gt_csv_paths) == 0:
        return result

    mAP_3d = AtekObb3Metrics(
        class_metrics=compute_per_class_metrics,
        global_name_to_id={
            name: int(sem_id) for sem_id, name in sem_id_to_name.items()
        },
    )

    # Loop over all pred and gt csv file pairs
    logger.info("loading csv files ... ")
    all_precisions = []
    all_recalls = []
    for pred_csv, gt_csv in zip(pred_csv_paths, gt_csv_paths):
        # Load in prediction and gt obbs as a {timestamp -> Obb3GtDict} dict
        pred_reader = AtekObb3CsvReader(input_filename=pred_csv)
        pred_obb_dict = pred_reader.read_as_obb_dict()
        gt_reader = AtekObb3CsvReader(input_filename=gt_csv)
        gt_obb_dict = gt_reader.read_as_obb_dict()

        # Update mAP metrics from this sequence
        single_sequence_result, per_seq_prec, pre_seq_recall = (
            update_from_single_sequence_obb3(
                pred_obb_dict=pred_obb_dict,
                gt_obb_dict=gt_obb_dict,
                mAP_3d=mAP_3d,
                iou=iou,
                confidence_lower_threshold=confidence_lower_threshold,
            )
        )

        # add prec recall to the list
        all_precisions.extend(per_seq_prec)
        all_recalls.extend(pre_seq_recall)

    # Compute mAP metrics numbers, ignore average recall (e.g. "mar_*")
    logger.info(f"computing mAP metrics numbers ... ")
    result_map = mAP_3d.compute()
    # ignore average recall (e.g. "rgb/mar_220_3D")
    result_map = {
        k: v.item() for k, v in result_map.items() if not k.startswith("mar_")
    }
    result.update(result_map)

    # Compute mean precision and recall
    result[f"precision@IoU{iou},Confidence{confidence_lower_threshold}"] = np.mean(
        all_precisions
    )
    result[f"recall@IoU{iou},Cofidence{confidence_lower_threshold}"] = np.mean(
        all_recalls
    )
    return result



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from dataclasses import asdict, dataclass
from typing import Optional

import numpy as np
from atek_v1.data_preprocess.data_schema import FramesetGroup
from atek_v1.data_preprocess.data_utils import (
    check_all_same_member,
    generate_disjoint_colors,
    unify_object_target,
)
from atek_v1.data_preprocess.frameset_aligner import FramesetAligner

# from atek_v1.utils import mesh_boolean_utils

from projectaria_tools.core.sophus import SE3

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@dataclass
class FramesetSelectionConfig:
    """
    Config class for selecting framesets. The basic idea to group frameset is based on
    the temporal information combined with the spatial information. For example, the simplest
    way to group frameset is starting from every frameset in order and find n framesets to group
    them, aka, group n framesets starting from first frameset with stride of 1.
    """

    # The number of framesets to group together.
    num_framesets_per_group: int = 3

    # Skip first n framesets in the frameset aligner.
    skip_first_n_framesets: int = 0

    # Skip last n framesets in the frameset aligner.
    skip_last_n_framesets: int = 0

    # The stride between two consecutive framesets group first frameset.
    stride: int = 1

    # Time threshold for selecting framesets. If set to None, we don't do any timestamp
    # thresholding for selection frameset for grouping.
    time_duration_ns_threshold: Optional[int] = None

    # Distance threshold for selecting framesets. If set to None, we don't do any translation
    # thresholding for selection frameset for grouping.
    translation_m_threshold: Optional[float] = None

    # Rotation threshold for selecting framesets. If set to None, we don't do any rotation
    # thresholding for selection frameset for grouping.
    rotation_deg_threshold: Optional[float] = None

    # Fov overlapping ratio threshold for selecting framesets. If set to None, we don't do any
    # overlapping ratio thresholding for selection frameset for grouping. This mechanism is that build
    # a fov for depth fov near clipping distance to fov far clipping distance, and compute
    # the overlapping ratio between two framesets to make sure the viewing perspective have changed enough.
    # We only select a frameset when the overlapping ratio is getting smaller than this threshold.
    # CAUTION: The compute of the mesh intersection is expensive and slow so only use this when necessary.
    fov_overlapping_ratio_threshold: Optional[float] = None

    # The near clipping distance for computing the fov overlapping ratio.
    far_clipping_distance: Optional[float] = 4.0

    # The frameset idx which selected as the local coordinate frame.
    local_selection: int = 0

    def __str__(self):
        return "FramesetSelectionConfig:\n" + "\n".join(
            [f"  {k}: {v}" for k, v in asdict(self).items()]
        )


class FramesetGroupGenerator:
    def __init__(
        self,
        frameset_aligner: FramesetAligner,
        frameset_selection_config: FramesetSelectionConfig,
        require_objects: bool = False,
    ):
        self.frameset_aligner = frameset_aligner
        self.frameset_selection_config = frameset_selection_config
        self.require_objects = require_objects

        self.sanity_check_inputs()
        self.frameset_ids_for_groups = self.process_framesets()

    def sanity_check_inputs(self):
        """
        Sanity check frameset grouping settings and inputs.
        """
        if (
            (self.frameset_selection_config.translation_m_threshold is not None)
            or (self.frameset_selection_config.rotation_deg_threshold is not None)
            or (
                self.frameset_selection_config.fov_overlapping_ratio_threshold
                is not None
            )
        ):
            assert (
                self.frameset_aligner.mps_data_processor is not None
            ), "Require trajectory information for frameset selection."

        assert self.frameset_selection_config.stride >= 1

    def process_framesets(self):
        """
        Process the framesets for grouping.
        """
        if self.frameset_selection_config.fov_overlapping_ratio_threshold is not None:
            self.frameset_aligner.update_frameset_fov_mesh(
                far_clipping_distance=self.frameset_selection_config.far_clipping_distance
            )

        num_framesets = self.frameset_aligner.aligned_frameset_number()
        frameset_ids_for_groups = []

        logger.info(
            f"Start grouping framesets based on the settings:\n {self.frameset_selection_config}"
        )

        for i in range(
            self.frameset_selection_config.skip_first_n_framesets,
            num_framesets - self.frameset_selection_config.skip_last_n_framesets,
            self.frameset_selection_config.stride,
        ):
            frameset_group_ids = self.group_framesets_by_setting(i)
            if frameset_group_ids is not None:
                frameset_ids_for_groups.append(frameset_group_ids)

        logger.info(
            f"Framesets grouping finished with {len(frameset_ids_for_groups)} "
            f"groups from {num_framesets} aligned framesets."
        )

        return frameset_ids_for_groups

    def get_frameset_selection_info_by_index(self, index: int):
        """
        Get the selected frameset info by index.
        """
        info = {}
        if self.frameset_selection_config.time_duration_ns_threshold is not None:
            info["frameset_timestamp_ns"] = (
                self.frameset_aligner.get_frameset_timestamp_by_index(index)
            )

        if (self.frameset_selection_config.translation_m_threshold is not None) or (
            self.frameset_selection_config.rotation_deg_threshold is not None
        ):
            info["T_world_frameset"] = (
                self.frameset_aligner.get_T_world_frameset_by_index(index)
            )

        if self.frameset_selection_config.fov_overlapping_ratio_threshold is not None:
            frameset_fov_in_world = (
                self.frameset_aligner.get_frameset_fov_mesh()
            ).apply_transform(
                self.frameset_aligner.get_T_world_frameset_by_index(index).to_matrix()
            )
            info["frameset_fov_in_world"] = frameset_fov_in_world

        return info

    def group_framesets_by_setting(
        self,
        start_index: int,
    ):
        """
        Group framesets by setting starting from the start_index frameset.
        """
        num_framesets = (
            self.frameset_aligner.aligned_frameset_number()
            - self.frameset_selection_config.skip_last_n_framesets
        )
        assert (
            0 <= start_index < num_framesets
        ), f"Start index {start_index} | {num_framesets}"

        if (
            (self.frameset_selection_config.time_duration_ns_threshold is None)
            and (self.frameset_selection_config.translation_m_threshold is None)
            and (self.frameset_selection_config.rotation_deg_threshold is None)
            and (self.frameset_selection_config.fov_overlapping_ratio_threshold is None)
        ):
            # Just try to get the consecutive framesets if we don't need to check any
            # grouping criteria.

            # Quick return if not enough framesets left to group.
            if (
                num_framesets - start_index - 1
                < self.frameset_selection_config.num_framesets_per_group - 1
            ):
                return None
            else:
                return list(
                    range(
                        start_index,
                        start_index
                        + self.frameset_selection_config.num_framesets_per_group,
                    )
                )

        frameset_group_ids = [start_index]
        last_frameset_info = self.get_frameset_selection_info_by_index(start_index)
        for i in range(start_index + 1, num_framesets):
            if (
                len(frameset_group_ids)
                == self.frameset_selection_config.num_framesets_per_group
            ):
                return frameset_group_ids

            # Quick return if not enough framesets left to group.
            if (
                num_framesets - i
                < self.frameset_selection_config.num_framesets_per_group
                - len(frameset_group_ids)
            ):
                return None

            current_frameset_info = self.get_frameset_selection_info_by_index(i)
            if self.frameset_selection_config.time_duration_ns_threshold is not None:
                if (
                    current_frameset_info["frameset_timestamp_ns"]
                    - last_frameset_info["frameset_timestamp_ns"]
                ) >= self.frameset_selection_config.time_duration_ns_threshold:
                    frameset_group_ids.append(i)
                    last_frameset_info = current_frameset_info
                    continue

            if self.frameset_selection_config.translation_m_threshold is not None:
                distance = np.linalg.norm(
                    current_frameset_info["T_world_frameset"].translation()
                    - last_frameset_info["T_world_frameset"].translation()
                )
                if distance >= self.frameset_selection_config.translation_m_threshold:
                    frameset_group_ids.append(i)
                    last_frameset_info = current_frameset_info
                    continue

            if self.frameset_selection_config.rotation_deg_threshold is not None:
                R = (
                    last_frameset_info["T_world_frameset"].rotation().inverse()
                    @ current_frameset_info["T_world_frameset"].rotation()
                ).to_matrix()
                trace = np.trace(R)
                theta = np.arccos((trace - 1) / 2) * 180 / np.pi
                if theta >= self.frameset_selection_config.rotation_deg_threshold:
                    frameset_group_ids.append(i)
                    last_frameset_info = current_frameset_info
                    continue

            if (
                self.frameset_selection_config.fov_overlapping_ratio_threshold
                is not None
            ):
                # TODO: temporarily disabled to pass CI. The preprocessing part of atek_v1 should NOT be used!
                """
                fov_intersection = mesh_boolean_utils.intersect_meshes(
                    last_frameset_info["frameset_fov_in_world"],
                    current_frameset_info["frameset_fov_in_world"],
                )
                if (
                    fov_intersection.volume
                    / last_frameset_info["frameset_fov_in_world"].volume
                    <= self.frameset_selection_config.fov_overlapping_ratio_threshold
                ):
                    frameset_group_ids.append(i)
                    last_frameset_info = current_frameset_info
                    continue
                """
                pass

        return None

    def frameset_group_number(self):
        return len(self.frameset_ids_for_groups)

    def get_frameset_group_by_index(self, index: int) -> FramesetGroup:
        assert 0 <= index < self.frameset_group_number(), f"Index {index} out of bound."
        frameset_group = FramesetGroup()
        frameset_group.framesets = [
            self.frameset_aligner.get_frameset_by_index(frameset_id)
            for frameset_id in self.frameset_ids_for_groups[index]
        ]

        # Fill in the frameset and some sanity checks
        assert (
            len(frameset_group.framesets) > 0
        ), "No valid framesets found for frameset group."
        assert check_all_same_member(
            frameset_group.framesets, "data_source"
        ), "All the frames must be from the same data source."
        check_all_same_member(
            frameset_group.framesets, "sequence_name"
        ), "All the frames must be from the same sequence."

        frameset_group.data_source = frameset_group.framesets[0].data_source
        frameset_group.sequence_name = frameset_group.framesets[0].sequence_name
        frameset_group.local_selection = self.frameset_selection_config.local_selection

        # Generate trajectory information if available.
        if frameset_group.framesets[0].T_world_frameset is not None:
            Ts_world_frameset = [
                SE3.from_matrix3x4(frameset.T_world_frameset)
                for frameset in frameset_group.framesets
            ]
            T_world_local = Ts_world_frameset[
                self.frameset_selection_config.local_selection
            ]
            T_local_world = T_world_local.inverse()
            frameset_group.Ts_local_frameset = [
                (T_local_world @ T_world_frameset).to_matrix3x4()
                for T_world_frameset in Ts_world_frameset
            ]
            frameset_group.T_world_local = T_world_local.to_matrix3x4()

            frameset_group.gravity_in_world = np.column_stack(
                [frameset.gravity_in_world for frameset in frameset_group.framesets]
            ).mean(axis=1, keepdims=True)
            frameset_group.gravity_in_local = (
                T_local_world.rotation() @ frameset_group.gravity_in_world
            )
            if self.require_objects:
                unify_object_target(frameset_group)
                frameset_group.Ts_local_object = [
                    (T_local_world @ SE3.from_matrix3x4(T_world_object)).to_matrix3x4()
                    for T_world_object in frameset_group.Ts_world_object
                ]

        return frameset_group

    def get_frameset_group_fov_mesh_by_index(self, index: int):
        """
        Return frameset group FOV mesh in the world coordinate.
        """
        assert 0 <= index < self.frameset_group_number(), f"Index {index} out of bound."
        frameset_meshes = []
        colors = generate_disjoint_colors(
            self.frameset_selection_config.num_framesets_per_group
        )
        for i, frameset_id in enumerate(self.frameset_ids_for_groups[index]):
            frameset_mesh = self.frameset_aligner.get_frameset_fov_mesh()
            T_world_frameset = self.frameset_aligner.get_T_world_frameset_by_index(
                frameset_id
            )
            frameset_mesh.apply_transform(T_world_frameset.to_matrix())
            frameset_mesh.visual.vertex_colors = colors[i]
            frameset_meshes.append(frameset_mesh)
        return sum(frameset_meshes)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from dataclasses import dataclass, fields
from typing import Dict, List, Optional

import numpy as np
import torch


@dataclass
class Frame:
    """
    A class to store the information of a frame in a video sequence.
    """

    # ------------------------ ARIA IMAGE DATA BEGIN -------------------------
    # The index of the frame in the vrs.
    frame_id: int = None

    # The sequence name of the vrs.
    sequence_name: str = None

    # The dataset name
    data_source: str = None

    # Camera model. e.g. [Pinhole, Fisheye, etc]
    camera_model: str = None

    # Camera model intrinsics parameters.
    camera_parameters: np.ndarray = None

    # Stream id name.
    stream_id: str = None

    # Camera label name.
    camera_name: str = None

    # Timestamp of the frame.
    timestamp_ns: int = None

    # Image data in rgb order in numpy array. HxWxC
    image: np.ndarray = None
    # ------------------------ ARIA IMAGE DATA END ---------------------------

    # +++++++++++++++++++++++ MPS DATA BEGIN +++++++++++++++++++++++++++
    # Gravity direction in camera coordinate frame. 3x1
    gravity_in_camera: np.ndarray = None

    # Gravity direction in world coordinate frame. 3x1
    gravity_in_world: np.ndarray = None

    # Transformation from camera to world coordinate frame. 3x4 [R|t]
    T_world_camera: np.ndarray = None

    # +++++++++++++++++++++++ MPS DATA END +++++++++++++++++++++++++++++

    # ======================= GROUND-TRUTH DATA BEGIN ===============================
    # Mapping for category id to name.
    category_id_to_name: Dict[int, str] = None

    # Object instances ids.
    object_instance_ids: List[int] = None

    # Object category semantic ids. Make sure this is 1 to 1 aligned with the instance ids.
    object_category_ids: List[int] = None

    # Transformation form object to camera for each object instance. List[ 3x4 [R|t] ]
    Ts_camera_object: List[np.ndarray] = None

    # Transformation from object to world for each object instance. List[ 3x4 [R|t] ]
    Ts_world_object: List[np.ndarray] = None

    # Object bounding box dimensions in meters. List[ 3 [xyz] ]
    object_dimensions: List[np.ndarray] = None

    # Object 2d bounding boxes in pixels. List[ [xmin, xmax, ymin, ymax] ]
    bb2ds: List[np.ndarray] = None
    # ======================= TARGET DATA END =================================

    @staticmethod
    def image_fields():
        return [
            "image",
        ]

    @staticmethod
    def camera_info_fields():
        return [
            "camera_model",
            "camera_name",
        ]

    @staticmethod
    def camera_data_fields():
        return [
            "camera_parameters",
        ]

    @staticmethod
    def input_info_fields():
        return [
            "frame_id",
            "sequence_name",
            "data_source",
            "stream_id",
            "timestamp_ns",
        ]

    @staticmethod
    def trajectory_data_fields():
        return ["gravity_in_camera", "gravity_in_world", "T_world_camera"]

    @staticmethod
    def object_info_fields():
        return [
            "category_id_to_name",
            "object_instance_ids",
            "object_category_ids",
        ]

    @staticmethod
    def object_2d_data_fields():
        return [
            "bb2ds",
        ]

    @staticmethod
    def object_3d_data_fields():
        return [
            "Ts_camera_object",
            "Ts_world_object",
            "object_dimensions",
        ]

    @staticmethod
    def flatten_dict_key_pattern():
        return re.compile(r"F#\d+-\d+\+.+")

    def to_flatten_dict(self):
        """
        Transforms the frame attributes into a flattened dictionary, excluding
        attributes with None values. Attributes are prefixed to ensure uniqueness and
        to maintain context.
        """
        flatten_dict = {
            f"F#{self.stream_id}+{f.name}": getattr(self, f.name)
            for f in fields(self)
            if getattr(self, f.name) is not None
        }
        return flatten_dict


@dataclass
class Frameset:
    """
    A class to store the information of a set of frames in a video sequence.
    """

    # ------------------------ INPUT IMAGE DATA BEGIN -------------------------
    # The frames in the frame set.
    frames: List[Frame] = None

    # The sequence name of the vrs.
    sequence_name: str = None

    # The dataset name
    data_source: str = None

    # Timestamp of the frameset which could be configured to different settings.
    # For example, it could be the average timestamp of the frames' time stamp.
    # It could also be the timestamp of the first frame in the frameset. Could
    # be extended base on the needs.
    timestamp_ns: int = None

    # Timestamp type. [Average, Earliest, ...]
    timestamp_type: str = None

    # Frameset rig origin selection method. Different ways to do that, could be set to
    # one camera, some virtual coordinate, or some other methods. Could be extended base
    # on the needs. [camera_name, stream_name, etc]
    origin_selection: str = None

    # Transformation from frameset to camera coordinate frame. 3x4 [R|t]
    # Note this is aligned with the frame list.
    Ts_frameset_camera: List[np.ndarray] = None
    # ------------------------ INPUT IMAGE DATA END ---------------------------

    # +++++++++++++++++++++++ MPS DATA BEGIN +++++++++++++++++++++++++++

    # Gravity direction in frameset coordinate frame. 3x1
    gravity_in_frameset: np.ndarray = None

    # Gravity direction in world coordinate frame. 3x1
    gravity_in_world: np.ndarray = None

    # Transformation from frameset to world coordinate frame. 3x4 [R|t]
    T_world_frameset: np.ndarray = None

    # Semi-dense point cloud, points in world coordinate frame. Nx3
    points_world: Optional[torch.Tensor] = None

    # Semi-dense point cloud, inverse points distance std, Nx1
    points_inv_dist_std: Optional[torch.Tensor] = None

    # +++++++++++++++++++++++ MPS DATA END +++++++++++++++++++++++++++++

    # ======================= TARGET DATA BEGIN ===============================
    # Frameset target is in the context for the frameset. For example, all the objects
    # seen by any frame in the frameset are considered as the target of the frameset.

    # Mapping for category id to name.
    category_id_to_name: Dict[int, str] = None

    # Object instances ids.
    object_instance_ids: List[int] = None

    # Object category semantic ids. Make sure this is 1 to 1 aligned with the instance ids.
    object_category_ids: List[int] = None

    # Transformation from object to world for each object instance. List[ 3x4 [R|t] ]
    Ts_world_object: List[np.ndarray] = None

    # Object bounding box dimensions in meters. List[ 3 [xyz] ]
    object_dimensions: List[np.ndarray] = None

    # Transformation form object to frame for each object instance. List[ 3x4 [R|t] ]
    Ts_frameset_object: List[np.ndarray] = None

    # ======================= TARGET DATA END =================================
    @staticmethod
    def input_info_fields():
        return [
            "sequence_name",
            "data_source",
            "timestamp_ns",
            "timestamp_type",
            "origin_selection",
        ]

    @staticmethod
    def camera_data_fields():
        return [
            "Ts_frameset_camera",
        ]

    @staticmethod
    def trajectory_data_fields():
        return ["gravity_in_frameset", "gravity_in_world", "T_world_frameset"]

    @staticmethod
    def semidense_points_fields():
        return [
            "points_world",
            "points_inv_dist_std",
        ]

    @staticmethod
    def object_info_fields():
        return [
            "category_id_to_name",
            "object_instance_ids",
            "object_category_ids",
        ]

    @staticmethod
    def object_3d_data_fields():
        return [
            "Ts_frameset_object",
            "Ts_world_object",
            "object_dimensions",
        ]

    @staticmethod
    def flatten_dict_key_pattern():
        return re.compile(r"FS\+.+")

    def to_flatten_dict(self):
        """
        Transforms the frameset's attributes into a flattened dictionary, excluding
        attributes with None values. The nested Frame structure is not preserved in
        the resulting dictionary. Attributes are prefixed to ensure uniqueness and
        to maintain context.
        """
        flatten_dict = {}
        for frame in self.frames:
            flatten_dict.update(frame.to_flatten_dict())

        for f in fields(self):
            if f.name == "frames":
                continue
            if getattr(self, f.name) is not None:
                flatten_dict[f"FS+{f.name}"] = getattr(self, f.name)

        return flatten_dict


@dataclass
class FramesetGroup:
    """
    A class to store the information of a group of framesets in a video sequence.
    """

    # ------------------------ INPUT IMAGE DATA BEGIN -------------------------
    # The framesets in the frameset group.
    framesets: List[Frameset] = None

    # The sequence name of the vrs.
    sequence_name: str = None

    # The dataset name
    data_source: str = None

    # Local coordinate frame selection method. Pick the ith frameset rig coordinate
    # as the local coordinate frame.
    local_selection: int = None
    # ------------------------ INPUT IMAGE DATA END ---------------------------

    # +++++++++++++++++++++++ TRAJECTORY DATA BEGIN +++++++++++++++++++++++++++
    # Gravity direction in frameset coordinate frame. 3x1
    gravity_in_local: np.ndarray = None

    # Gravity direction in world coordinate frame. 3x1
    gravity_in_world: np.ndarray = None

    # Transformation from local to world coordinate frames. 3x4 [R|t]
    T_world_local: np.ndarray = None

    # Transformation from local to frameset coordinate frames. 3x4 [R|t]
    # Note this is aligned with the frameset list.
    Ts_local_frameset: List[np.ndarray] = None
    # +++++++++++++++++++++++ TRAJECTORY DATA END +++++++++++++++++++++++++++++

    # ======================= TARGET DATA BEGIN ===============================
    # Frameset target is in the context for the frameset. For example, all the objects
    # seen by any frame in the frameset are considered as the target of the frameset.

    # Mapping for category id to name.
    category_id_to_name: Dict[int, str] = None

    # Object instances ids.
    object_instance_ids: List[int] = None

    # Object category semantic ids. Make sure this is 1 to 1 aligned with the instance ids.
    object_category_ids: List[int] = None

    # Transformation from object to world for each object instance. List[ 3x4 [R|t] ]
    Ts_world_object: List[np.ndarray] = None

    # Object bounding box dimensions in meters. List[ 3 [xyz] ]
    object_dimensions: List[np.ndarray] = None

    # Transformation form object to frame for each object instance. List[ 3x4 [R|t] ]
    Ts_local_object: List[np.ndarray] = None
    # ======================= TARGET DATA END =================================

    @staticmethod
    def input_info_fields():
        return [
            "sequence_name",
            "data_source",
            "local_selection",
        ]

    @staticmethod
    def trajectory_data_fields():
        return [
            "gravity_in_local",
            "gravity_in_world",
            "T_world_local",
            "Ts_local_frameset",
        ]

    @staticmethod
    def object_info_fields():
        return [
            "category_id_to_name",
            "object_instance_ids",
            "object_category_ids",
        ]

    @staticmethod
    def object_3d_data_fields():
        return [
            "Ts_local_object",
            "Ts_world_object",
            "object_dimensions",
        ]

    @staticmethod
    def flatten_dict_key_pattern():
        return re.compile(r"FSG\+.+")

    def to_flatten_dict(self):
        """
        Transforms the FramesetGroup's attributes into a flattened dictionary, excluding
        attributes with None values. The nested Frame/Frameset structure is not preserved in
        the resulting dictionary. Attributes are prefixed to ensure uniqueness and
        to maintain context.
        """
        flatten_dict = {}
        for i, frameset in enumerate(self.framesets):
            frameset_flatten_dict = frameset.to_flatten_dict()
            for k, v in frameset_flatten_dict.items():
                if i == 0:
                    flatten_dict[k] = [v]
                else:
                    flatten_dict[k].append(v)

        for f in fields(self):
            if f.name == "framesets":
                continue
            if getattr(self, f.name) is not None:
                flatten_dict[f"FSG+{f.name}"] = getattr(self, f.name)

        return flatten_dict



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from argparse import Namespace
from typing import Dict

from cubercnn.config import get_cfg_defaults
from cubercnn.modeling.backbone import build_dla_from_vision_fpn_backbone  # noqa
from cubercnn.modeling.meta_arch import build_model, RCNN3D  # noqa
from cubercnn.modeling.proposal_generator import RPNWithIgnore  # noqa
from cubercnn.modeling.roi_heads import ROIHeads3D  # noqa
from detectron2.checkpoint import DetectionCheckpointer
from detectron2.config import get_cfg
from detectron2.engine import default_setup


def create_cubercnn_config(args: Namespace) -> Dict:
    """
    Create configs and perform basic setups.
    """
    assert args.model_name == "cubercnn"
    cfg = get_cfg()
    get_cfg_defaults(cfg)

    # add extra configs for data
    cfg.MAX_TRAINING_ATTEMPTS = 3
    cfg.TRAIN_LIST = ""
    cfg.TEST_LIST = ""
    cfg.ID_MAP_JSON = ""
    cfg.OBJ_PROP_JSON = ""
    cfg.CATEGORY_JSON = ""
    cfg.DATASETS.OBJECT_DETECTION_MODE = ""
    cfg.SOLVER.VAL_MAX_ITER = 0

    cfg.merge_from_file(args.config_file)
    if "opts" in args:
        cfg.merge_from_list(args.opts)

    cfg.freeze()
    default_setup(cfg, args)

    model_config = {
        "ckpt_dir": os.path.dirname(args.config_file),
        "cfg": cfg,
        "score_threshold": args.threshold,
        "category_names": cfg.DATASETS.CATEGORY_NAMES,
    }

    return model_config


def create_cubercnn_model(model_config: Dict):
    """
    Create CubeRCNN model for inference from config
    """
    model = build_model(model_config["cfg"], priors=None)
    _ = DetectionCheckpointer(model, save_dir=model_config["ckpt_dir"]).resume_or_load(
        model_config["cfg"].MODEL.WEIGHTS, resume=True
    )
    model.eval()

    return model



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# atek/data_preprocess/__init__.py

from .adt_gt_data_processor import AdtGtDataProcessor  # noqa
from .frame_data_processor import FrameDataProcessor  # noqa
from .frameset_aligner import FramesetAligner  # noqa
from .frameset_group_generator import (  # noqa
    FramesetGroupGenerator,
    FramesetSelectionConfig,
)
from .mps_data_processor import MpsDataProcessor  # noqa
from .webdataset_writer import AtekWdsWriter, DataSelectionSettings  # noqa



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import numpy as np


def compute_precision_recall_curve(
    y_true: List[bool], y_score: List[float], total_gt_count: int
):
    """Compute precision-recall pairs for different probability thresholds
    This function implementation is based on the scikit-learn
    precision_recall_curve function implementation.
    But it fixes the issue that the scikit-learn function cannot properly
    handle false negatives.

    Parameters
    ----------
    y_true : array, shape = [n_samples]
        True targets of binary classification

    y_score : array, shape = [n_samples]
        Estimated probabilities or decision function

    total_gt_count : int
        Total number of ground-truth samples

    Returns
    -------
    precision : array, shape = [n_thresholds + 1]
        Precision values such that element i is the precision of
        predictions with score >= thresholds[i] and the last element is 1.

    recall : array, shape = [n_thresholds + 1]
        Decreasing recall values such that element i is the recall of
        predictions with score >= thresholds[i] and the last element is 0.

    thresholds : array, shape = [n_thresholds <= len(np.unique(y_score))]
        Increasing thresholds on the decision function used to compute
        precision and recall.
    """

    if len(y_true) != len(y_score):
        raise ValueError("y_true and y_score should have the same length. ")
    if len(y_true) == 0:
        raise ValueError("y_true or y_score should not be empty.")
    if total_gt_count <= 0:
        raise ValueError("total_gt_count should be positive.")

    # sort scores and corresponding truth values
    desc_score_indices = np.argsort(y_score, kind="mergesort")[::-1]
    y_score = y_score[desc_score_indices]
    y_true = y_true[desc_score_indices]

    # y_score typically has many tied values. Here we extract
    # the indices associated with the distinct values. We also
    # concatenate a value for the end of the curve.
    distinct_value_indices = np.where(np.diff(y_score))[0]
    threshold_idxs = np.r_[distinct_value_indices, y_true.size - 1]

    # accumulate the true positives with decreasing threshold
    tps = np.cumsum(y_true)[threshold_idxs]
    # false positives with decreasing threshold
    fps = 1 + threshold_idxs - tps
    thresholds = y_score[threshold_idxs]

    positives = tps + fps
    precision = np.divide(tps, positives, where=(positives != 0))
    recall = tps / total_gt_count

    return precision, recall, thresholds


def compute_average_precision(
    y_true: List[bool], y_score: List[float], total_gt_count: int
):
    """Compute average precision (AP) from prediction scores

    This function implementation is based on the scikit-learn
    precision_recall_curve function implementation.
    But it fixes the issue that the scikit-learn function cannot properly
    handle false negatives.

    Parameters
    ----------
    y_true : array, shape = [n_samples]
        True targets of binary classification

    y_score : array, shape = [n_samples]
        Estimated probabilities or decision function

    total_gt_count : int
        Total number of ground-truth samples

    Returns
    -------
    average_precision : float

    References
    ----------
    .. [1] `Wikipedia entry for the Average precision
           <https://en.wikipedia.org/w/index.php?title=Information_retrieval&
           oldid=793358396#Average_precision>`_
    """
    # precision is sorted in decreasing order, recall is sorted in increasing order
    precision, recall, _ = compute_precision_recall_curve(
        y_true, y_score, total_gt_count
    )
    # Add the precision = 1.0, recall = 0.0 starting point to compute the AP
    precision = np.r_[1.0, precision]
    recall = np.r_[0.0, recall]

    # Apply filter to precisions to make sure the precision is no worse than precision
    # value when the recall is higher (otherwise, we can always pick the confidence
    # threshold that achieve both better precision and recall)
    precision = np.maximum.accumulate(precision[::-1])[::-1]

    # Return the step function integral
    return np.sum(np.diff(recall) * np.array(precision)[1:])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from abc import ABC, abstractmethod
from typing import Optional

from atek_v1.data_preprocess.data_schema import Frame

from projectaria_tools.core import calibration

from projectaria_tools.core.sophus import SE3
from projectaria_tools.core.stream_id import StreamId

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class BaseGtDataProcessor(ABC):
    def __init__(
        self,
        name: str,
        stream_id: StreamId,
    ):
        """
        name: data type readable name
        """
        self.name = name
        self.data_source = "Unknown"
        self.stream_id = stream_id

    @abstractmethod
    def set_undistortion_params(
        self,
        original_camera_calibration: calibration.CameraCalibration,
        target_camera_calibration: Optional[calibration.CameraCalibration] = None,
        rotate_image_cw90deg: bool = True,
    ):
        pass

    @abstractmethod
    def get_object_gt_at_timestamp_ns(
        self,
        frame: Frame,
        timestamp_ns: int,
        T_camera_world: SE3,
        stream_id: Optional[StreamId] = None,
        tolerance_ns: int = 1000_000,
    ):
        pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from collections import defaultdict
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd

import torch
from atek.data_preprocess.atek_data_sample import MpsSemiDensePointData

from omegaconf.omegaconf import DictConfig

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

TIMESTAMP_KEY_IN_DF: str = "tracking_timestamp_us"


class MpsSemiDenseProcessor:
    def __init__(
        self,
        mps_semidense_points_file: str,
        mps_semidense_observations_file: str,
        conf: DictConfig,
    ):
        # Parse in conf
        self.conf = conf

        # Load in semidense points data. Not using MPSDataProvider because it is not sufficient.
        time_0 = time.time()
        self.uid_to_p3, self.uid_to_dist_std, self.uid_to_inv_dist_std = (
            self._load_semidense_global_points(mps_semidense_points_file)
        )
        time_1 = time.time()
        self.time_to_uids, self.uid_to_times = self._load_semidense_observations(
            mps_semidense_observations_file
        )
        time_2 = time.time()

        self._compute_semidense_volume()
        time_3 = time.time()

        logger.info(
            f"loading semidense points takes {time_1-time_0} seconds, observations takes {time_2-time_1} seconds"
            f"and computing semidense volume takes {time_3-time_2} seconds"
        )

    def get_semidense_points_by_timestamps_ns(
        self,
        timestamps_ns: List[int],
    ) -> Optional[MpsSemiDensePointData]:
        """
        Obtain a semidense points data by timestamp, where `points_world` are in meters, and `points_inv_dist_std` are in meter^-1.
        returns: if successful, returns (points_world: List[torch.Tensor (N,3)], points_inv_dist_std: List[torch.Tensor (N)]), where len(List) = number of frames, which is 1
                else returns None
        """
        # create data frame to re-use timestamp matching function
        time_to_uids_df = pd.DataFrame(
            list(self.time_to_uids.items()), columns=["tracking_timestamp_us", "uids"]
        )

        matched_time_to_uids_df = self._find_matching_timestamps_in_df(
            data_frame=time_to_uids_df,
            timestamps_ns=timestamps_ns,
            tolerance_ns=self.conf.tolerance_ns,
            only_return_valid=False,
        )

        points_world_all = []
        dist_std_all = []
        inv_dist_std_all = []
        # loop over all matched timestamps, and stack point_in_world into a Nx3 tensor
        for uid_list in matched_time_to_uids_df["uids"]:
            # uid_list can also be a single nan value, indicating empty observations at this timestamp. hence needs to handle this separately
            if isinstance(uid_list, List):
                points_world = []
                dist_std = []
                inv_dist_std = []
                for uid in uid_list:
                    if (
                        (uid in self.uid_to_p3)
                        and (uid in self.uid_to_dist_std)
                        and (uid in self.uid_to_inv_dist_std)
                    ):
                        points_world.append(self.uid_to_p3[uid])
                        dist_std.append(self.uid_to_dist_std[uid])
                        inv_dist_std.append(self.uid_to_inv_dist_std[uid])
                    else:
                        raise ValueError(
                            f"Point UID {uid} not found in global semidense point file!"
                        )
                # Sort points by inv_distance, ascending
                combined = list(zip(inv_dist_std, points_world, dist_std))
                combined = sorted(combined, key=lambda x: x[0])
                inv_dist_std, points_world, dist_std = map(list, zip(*combined))

                # end for uid
                points_world_all.append(torch.stack(points_world, dim=0))
                dist_std_all.append(torch.tensor(dist_std))
                inv_dist_std_all.append(torch.tensor(inv_dist_std))
            else:
                points_world_all.append(torch.full((1, 3), float("nan")))
                dist_std_all.append(torch.tensor([float("nan")]))
                inv_dist_std_all.append(torch.tensor([float("nan")]))
        # end for uid_list

        capture_timestamps_ns = torch.tensor(
            matched_time_to_uids_df[TIMESTAMP_KEY_IN_DF].values * 1e3, dtype=torch.int64
        )

        return MpsSemiDensePointData(
            points_world=points_world_all,
            points_dist_std=dist_std_all,
            points_inv_dist_std=inv_dist_std_all,
            capture_timestamps_ns=capture_timestamps_ns,
            points_volumn_max=self.vol_max,
            points_volumn_min=self.vol_min,
        )

    def _load_semidense_global_points(
        self,
        path: str,
    ):
        logger.info(f"loading global semi-dense points from {path}")

        # Determine compression method
        if path.endswith(".csv"):
            compression_method = None
        elif path.endswith(".gz"):
            compression_method = "gzip"
        else:
            raise ValueError(f"Unsupported compression method for {path}")

        uid_to_p3 = {}
        uid_to_dist_std = {}
        uid_to_inv_dist_std = {}

        with open(path, "rb") as f:
            csv_data = pd.read_csv(f, compression=compression_method)

            # select points and uids and return mapping
            uid_pts = csv_data[
                ["uid", "dist_std", "inv_dist_std", "px_world", "py_world", "pz_world"]
            ]

            for row in uid_pts.values:
                uid = int(row[0])
                dist_std = float(row[1])
                inv_dist_std = float(row[2])
                p3 = torch.from_numpy(row[3:]).float()
                uid_to_p3[uid] = p3
                uid_to_dist_std[uid] = dist_std
                uid_to_inv_dist_std[uid] = inv_dist_std

        return uid_to_p3, uid_to_dist_std, uid_to_inv_dist_std

    def _load_semidense_observations(
        self,
        path: str,
    ):
        """
        Load semidense observations from a csv file, returns two-way mapping between timestamp_in_us and point uids.
        Args:
            path: The path to the csv file.
        Returns:
            A tuple of two dictionaries.
            The first dictionary maps from timestamp to a list of uids.
            The second dictionary maps from uid to a list of timestamps.
        """

        logger.info(f"loading semidense observations from {path}")

        # Determine compression method
        if path.endswith(".csv"):
            compression_method = None
        elif path.endswith(".gz"):
            compression_method = "gzip"
        else:
            raise ValueError(f"Unsupported compression method for {path}")

        time_to_uids = defaultdict(list)
        uid_to_times = defaultdict(list)

        with open(path, "rb") as f:
            csv = pd.read_csv(f, compression=compression_method)
            csv = csv[["uid", "frame_tracking_timestamp_us"]]
            for row in csv.values:
                uid = int(row[0])
                time_ns = int(row[1])
                time_to_uids[time_ns].append(uid)
                uid_to_times[uid].append(time_ns)
        return time_to_uids, uid_to_times

    def _find_matching_timestamps_in_df(
        self,
        data_frame: pd.DataFrame,
        timestamps_ns: Union[np.ndarray, List[int]],
        tolerance_ns: int,
        only_return_valid: bool = False,
    ) -> pd.DataFrame:
        """
        Helper function that given a list of timestamps, find the rows containing matching timestamps in the data frame, with some tolerance.
        Returns:
            a data frame containing the matching rows.
        """
        if isinstance(timestamps_ns, List):
            timestamps_ns = np.array(timestamps_ns)

        timestamps_us_df = pd.DataFrame(
            {TIMESTAMP_KEY_IN_DF: pd.Series(np.round(timestamps_ns / 1000).astype(int))}
        )
        timestamps_us_df = timestamps_us_df.sort_values(TIMESTAMP_KEY_IN_DF)

        # Use merge_asof to find the nearest pose for each timestamp.
        merged_df = pd.merge_asof(
            timestamps_us_df,
            data_frame,
            on=TIMESTAMP_KEY_IN_DF,
            tolerance=int(round(tolerance_ns / 1000)),
            direction="nearest",
        )

        if not only_return_valid:
            return merged_df
        else:
            # Check percentage of invalid timestamps.
            valid_merged_df = merged_df.dropna()

            invalid_count = len(merged_df) - len(valid_merged_df)
            invalid_percent = (invalid_count / len(merged_df)) * 100
            if invalid_percent > 5:
                logger.warning(
                    f"{invalid_count} ({invalid_percent:.2f}%) of the input timestamps can not find corresponding poses."
                )

            return valid_merged_df

    def _compute_semidense_volume(
        self, gpu_memory_mb=8000, quantiles=[0.001, 0.01, 0.05], voxel_size=0.04
    ):
        """
        compute the bounding volume for TSDF fusion

        gpu_memory: how much memory to be reserved for TSDF fusion, default: 8000MB
        quantiles: the quantiles of the point cloud in each 3d dimension to bound the global space,
        default: [0.001, 0.01, 0.05], will test which quantile is the smallest that can fit the memory
        voxel_size: the voxel size in meters for TSDF fusion, default: 0.04
        """
        # at least hold tsdf_vol, vol_weight, vol_color, vox_coords
        vol_memory = gpu_memory_mb / 4
        # assume float32 for volume dtype, how many voxels `vol_memory` translates to
        max_voxels = vol_memory * 1e6 / 4

        # Aggregate all global points
        all_points = torch.stack(list(self.uid_to_p3.values()))

        for q in quantiles:
            self.vol_min = torch.quantile(all_points, q, dim=0)
            self.vol_max = torch.quantile(all_points, 1 - q, dim=0)
            self.vol_min = self.vol_min.detach().cpu()
            self.vol_max = self.vol_max.detach().cpu()

            vox_dim = (self.vol_max - self.vol_min) / voxel_size
            est_num_voxels = vox_dim[0] * vox_dim[1] * vox_dim[2]
            if est_num_voxels < max_voxels:
                print(f"compute global bounding volume as {q} to {1-q} quantile")
                break
        if est_num_voxels > max_voxels:
            print("Warning: scene volume too large for TSDF fusion")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch


def get_cuboid_corners(half_extents: torch.Tensor) -> torch.Tensor:
    """
    https://github.com/facebookresearch/pytorch3d/blob/main/pytorch3d/ops/iou_box3d.py#L111)

    (4) +---------+. (5)
        | ` .     |  ` .
        | (0) +---+-----+ (1)
        |     |   |     |
    (7) +-----+---+. (6)|
        ` .   |     ` . |
        (3) ` +---------+ (2)
    NOTE: Throughout this implementation, we assume that boxes
    are defined by their 8 corners exactly in the order specified in the
    diagram above for the function to give correct results. In addition
    the vertices on each plane must be coplanar.
    As an alternative to the diagram, this is a unit bounding
    box which has the correct vertex ordering:

    box_corner_vertices = [
        [0, 0, 0],
        [1, 0, 0],
        [1, 1, 0],
        [0, 1, 0],
        [0, 0, 1],
        [1, 0, 1],
        [1, 1, 1],
        [0, 1, 1],
    ]

    Args:
        half_extents: Half extents of the cuboid in the format of (N, 3)

    Returns:
        corners: Corners of the cuboid in the format of (N, 8, 3)
    """
    corners = torch.tensor(
        [
            [-1, -1, -1],
            [1, -1, -1],
            [1, 1, -1],
            [-1, 1, -1],
            [-1, -1, 1],
            [1, -1, 1],
            [1, 1, 1],
            [-1, 1, 1],
        ],
        dtype=torch.float32,
        device=half_extents.device,
    )

    # Scale the corners of the unit box
    corners = corners * half_extents.unsqueeze(1)

    return corners


def batch_transform_points(
    points_in_B: torch.tensor, T_A_B: torch.Tensor
) -> torch.Tensor:
    """
    Return point_in_A = R_A_B @ points_in_B + t_A_B in shape [N x M x 3]
    Args:
        points_in_B (torch.Tensor): NxMx3 tensor for a batch N of M 3D points in frame B,
          corresponding to each transformation matrix
        T_A_B (torch.Tensor): Nx3x4 transformation matrices from B to A

    Returns:
        points_in_A (torch.Tensor): NxMx3 tensor of a batch N of M 3D points in frame A, after
            transformation
    """

    # Reshape points to (N, 3, M) for batch matrix multiplication
    points_in_B_reshaped = points_in_B.permute(0, 2, 1)
    M = points_in_B_reshaped.shape[-1]

    R_A_B = T_A_B[:, :, :3]
    t_A_B = T_A_B[:, :, 3]
    points_in_A = (
        torch.bmm(R_A_B, points_in_B_reshaped) + t_A_B.unsqueeze(-1).repeat(1, 1, M)
    ).permute(0, 2, 1)

    return points_in_A


def batch_inverse(T: torch.Tensor) -> torch.Tensor:
    """
    Inverse SE3 transformation matrix T in Tensor of shape (..., 3, 4).

    Args:
        T: (..., 3, 4) tensor, transformation matrix

    Returns:
        inversed_T: (..., 3, 4) tensor, inverse transformation matrix
    """
    inversed_T = T.clone()
    inversed_T[..., :, :3] = T[..., :, :3].transpose(1, 2)
    inversed_T[..., :, 3] = (
        -inversed_T[..., :, :3] @ (T[..., :, 3].unsqueeze(2))
    ).squeeze()
    return inversed_T



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List

import numpy as np
import pandas as pd
import torch

from atek_v1.evaluation.per_scene_metrics import (
    compute_per_scene_metrics,
    METRIC_DATA_TYPES,
)
from atek_v1.utils.eval_utils import compute_average_precision
from atek_v1.utils.obb3 import Obb3


def _set_false_positive(fp_df: pd.DataFrame) -> pd.DataFrame:
    false_positive_df = fp_df.copy()
    false_positive_df.loc[:, "gt_id"] = -1
    for col in false_positive_df.columns:
        if col not in [
            "category_id",
            "pred_id",
            "gt_id",
            "confidence",
        ]:
            false_positive_df.loc[:, col] = np.nan
    return false_positive_df


def _set_false_negative(fn_df: pd.DataFrame) -> pd.DataFrame:
    false_negative_df = fn_df.copy()
    false_negative_df.loc[:, "pred_id"] = -1
    for col in false_negative_df.columns:
        if col not in [
            "category_id",
            "pred_id",
            "gt_id",
        ]:
            false_negative_df.loc[:, col] = np.nan
    return false_negative_df


class Bbox3DEvaluator:
    """
    Evaluator for 3D bounding boxes
    """

    def __init__(
        self,
        metric_name: str = "IoU",
        metric_thresh: float = 0.25,
    ):
        """
        Args:
            metric_name (str): name of the metric used to match predicted and GT 3D boxes, such as IoU
                GIoU
            metric_thresh (float): threshold to decide if metric between predicted and GT is enough to
                be considered a good match.
        """
        self.metric_name = metric_name
        self.metric_thresh = metric_thresh
        if metric_name in ["IoU", "GIoU"]:
            self.metric_higher = True
        else:
            self.metric_higher = False
        self.matched_metrics_df_list = []

    def update(self, model_input: List[Dict], model_prediction: List[Dict]) -> None:
        """
        Compute 3D bounding box match metrics (IoU, GIoU, etc) for the model input and prediction
        pair, accumulate the results
        """
        for input, prediction in zip(model_input, model_prediction):
            if len(input["object_dimensions"]) == 0:
                obb3_gt = None
            else:
                dim_gt = input["object_dimensions"]
                t_world_obj_gt = input["Ts_world_object"][:, :, 3]
                R_world_obj_gt = input["Ts_world_object"][:, :, :3]
                category_id_gt = input["instances"].gt_classes
                fake_instance_id_gt = torch.Tensor(list(range(len(category_id_gt))))

                obb3_gt = Obb3(
                    dim_gt,
                    t_world_obj_gt,
                    R_world_obj_gt,
                    fake_instance_id_gt,
                    category_id_gt,
                )

            N_pred = len(prediction)
            if N_pred == 0:
                obb3_pred = None
            else:
                dim_pred = torch.Tensor([pred["dimensions"] for pred in prediction])
                t_world_obj_pred = torch.zeros((N_pred, 3))
                R_world_obj_pred = torch.zeros((N_pred, 3, 3))
                T_world_cam_4x4 = torch.eye(4)
                T_world_cam_4x4[:3, :] = input["T_world_camera"]
                for i, pred in enumerate(prediction):
                    T_cam_obj_pred_4x4 = torch.eye(4)
                    T_cam_obj_pred_4x4[:3, 3] = torch.Tensor(pred["t_cam_obj"])
                    T_cam_obj_pred_4x4[:3, :3] = torch.Tensor(pred["R_cam_obj"])
                    T_world_obj_pred_4x4 = T_world_cam_4x4 @ T_cam_obj_pred_4x4

                    t_world_obj_pred[i, :] = T_world_obj_pred_4x4[:3, 3]
                    R_world_obj_pred[i, :, :] = T_world_obj_pred_4x4[:3, :3]
                category_id_pred = torch.Tensor(
                    [pred["category_idx"] for pred in prediction]
                )
                fake_instance_id_pred = torch.Tensor(list(range(len(category_id_pred))))
                score_pred = torch.Tensor([pred["score"] for pred in prediction])
                obb3_pred = Obb3(
                    dim_pred,
                    t_world_obj_pred,
                    R_world_obj_pred,
                    fake_instance_id_pred,
                    category_id_pred,
                    score_pred,
                )

            # Compute per-scene (frame/frameset/frameset group) pairwise metrics (IoU/GIoU/Chamfer distance/etc)
            metrics_df = compute_per_scene_metrics(obb3_pred, obb3_gt)
            if len(metrics_df) == 0:
                continue
            # Match predicted boxes to GT boxes
            matched_metrics_df = self.match_pred_to_gt(metrics_df)

            matched_metrics_df.insert(0, "timestamp_ns", input["timestamp_ns"])

            # Append the matched metrics for the final evaluation aggregation
            self.matched_metrics_df_list.append(matched_metrics_df)

    def match_pred_to_gt(self, per_scene_metrics_df: pd.DataFrame) -> pd.DataFrame:
        """
        Match prediction 3D bounding boxes to ground truth 3D bounding boxes based on per-scene
        pairwise metrics, given a match criterion. For example, if the match criterion is IoU of
        threshold 0.25, then a predicted box with IoU over 0.25 to a GT box is considered a match. All
        the metrics will be returned for the matched pred and GT pair, such as IoU and GIoU.

        Args:
            per_scene_metrics_df (pd.DataFrame): DataFrame containing pred-to-GT metrics for a scene.

        Returns:
            matched_metrics_df (pd.DataFrame): DataFrame containing all metrics for each matched pred
                and GT boxes, and unmatched pred/GT boxes (metrics will be np.nan).
        """
        # Find false positives and false negatives
        false_positives = per_scene_metrics_df.query("gt_id == -1")
        false_negatives = per_scene_metrics_df.query("pred_id == -1")

        # Define condition for matching
        sign = ">=" if self.metric_higher else "<="
        match_condition = f"{self.metric_name} {sign} {self.metric_thresh}"

        # Process remaining predictions and GTs
        filtered_df = per_scene_metrics_df.query("gt_id != -1 & pred_id != -1")
        true_positives = []
        additional_false_positives = []
        additional_false_negatives = []

        # Append true positives
        matched_df = filtered_df.query(f"{match_condition}")

        if not matched_df.empty:
            if self.metric_higher:
                row_idx = matched_df.groupby("gt_id")[self.metric_name].idxmax()
            else:
                row_idx = matched_df.groupby("gt_id")[self.metric_name].idxmin()
            true_positives.append(matched_df.loc[row_idx])

        # Append additional false positives: preds not meeting match criterion
        matched_pred_ids = matched_df["pred_id"].unique().tolist()  # noqa
        curr_false_positives = filtered_df.query(
            "pred_id not in @matched_pred_ids"
        ).reset_index(drop=True)
        additional_false_positives.append(
            _set_false_positive(curr_false_positives).drop_duplicates()
        )

        # Append additional false negatives: GTs that have no valid matched preds,
        # i.e., preds with not meeting match criterion
        matched_gt_ids = matched_df["gt_id"].unique().tolist()  # noqa
        curr_false_negatives = filtered_df.query(
            "gt_id not in @matched_gt_ids"
        ).reset_index(drop=True)
        additional_false_negatives.append(
            _set_false_negative(curr_false_negatives).drop_duplicates()
        )

        if len(true_positives) > 0:
            true_positives = pd.concat(true_positives)
        if len(additional_false_positives) > 0:
            additional_false_positives = pd.concat(additional_false_positives)
        if len(additional_false_negatives) > 0:
            additional_false_negatives = pd.concat(additional_false_negatives)

        matched_metrics_df = [
            item
            for item in [
                false_positives,
                false_negatives,
                true_positives,
                additional_false_positives,
                additional_false_negatives,
            ]
            if len(item) > 0
        ]
        matched_metrics_df = (
            pd.concat(matched_metrics_df)
            .astype(METRIC_DATA_TYPES)
            .sort_values(by=["pred_id", "gt_id"])
            .reset_index(drop=True)
        )

        return matched_metrics_df

    def evaluate(self) -> Dict[str, Any]:
        """
        Use the accumulated results to run full evaluation
        """
        matched_metrics_df_all = pd.concat(self.matched_metrics_df_list)

        gt_count = matched_metrics_df_all.query("gt_id != -1").shape[0]
        if gt_count == 0:
            return {"mAP": np.nan}

        df = matched_metrics_df_all.query("pred_id != -1").sort_values(
            by=["confidence"], ascending=False
        )
        y_true = df["gt_id"].to_numpy() != -1
        if len(y_true) == 0:
            return {"mAP": 0}

        y_scores = df["confidence"].to_numpy()
        mean_ap = compute_average_precision(y_true, y_scores, gt_count)
        results = {"mAP": mean_ap}

        return results



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# pyre-strict

import logging
from typing import Optional

from atek.data_preprocess.atek_data_sample import AtekDataSample
from atek.data_preprocess.atek_wds_writer import AtekWdsWriter
from atek.viz.atek_visualizer import NativeAtekSampleVisualizer

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class GeneralAtekPreprocessor:
    """
    A base class that defines a high-level interface for preprocessing ATEK data.
    It consists of a subsampler to control subsampling, a sample builder to build samples, a WDS writer to write samples to disk, and a visualizer to visualize the results.
    """

    def __init__(
        self,
        sample_builder,  # Should be Union[ObbSampleBuilder, EfmSampleBuidler], but intentionally not specifying the type here for extensibility
        subsampler,  # Intentionally not specifying the type here for extensibility
        atek_wds_writer: Optional[AtekWdsWriter],
        atek_visualizer: Optional[NativeAtekSampleVisualizer],
    ) -> None:
        """
        init function
        """
        self.subsampler = subsampler
        self.sample_builder = sample_builder
        self.atek_wds_writer = atek_wds_writer
        self.atek_visualizer = atek_visualizer

        # TODO: maybe perform an API check for subsampler and sample_builder

    def __getitem__(self, index) -> Optional[AtekDataSample]:
        """
        API to get an AtekDataSample by index
        """
        timestamps_ns = self.subsampler.get_timestamps_by_sample_index(index)
        return self.sample_builder.get_sample_by_timestamps_ns(timestamps_ns)

    def process_all_samples(
        self, write_to_wds_flag: bool = True, viz_flag: bool = False
    ) -> int:
        """
        API to process all samples, and (optionally) write them to WDS and visualize them.
        Return the total number of valid samples being processed.
        """
        # Check if the WDS writer and visualizer are initialized
        if write_to_wds_flag:
            assert (
                self.atek_wds_writer is not None
            ), "AtekWdsWriter is not initialized, cannot write to WDS"
        if viz_flag:
            assert (
                self.atek_visualizer is not None
            ), "AtekVisualizer is not initialized, cannot visualize samples"

        # Loop over all samples, check for validity, and write them to WDS and visualize them if specified
        num_samples = 0
        for i in range(self.subsampler.get_total_num_samples()):
            sample = self.__getitem__(i)
            if sample is not None:
                num_samples += 1
                if write_to_wds_flag:
                    self.atek_wds_writer.add_sample(sample)
                if viz_flag:
                    self.atek_visualizer.plot_atek_sample(sample)

        if write_to_wds_flag:
            self.atek_wds_writer.close()
        if viz_flag:
            self.atek_visualizer.save_viz()

        logger.info(f"ATEK has processed {num_samples} valid samples in total.")
        return num_samples



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from collections import defaultdict

from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class MpsDataProcessor:
    def __init__(
        self,
        name: str,
        trajectory_file: str,
        semidense_global_point_file: Optional[str] = None,
        semidense_observation_file: Optional[str] = None,
        semidense_compression_method: Optional[str] = None,
    ):
        self.name = name
        # load in trajectory file
        self.trajectory_file = trajectory_file
        self.traj_df = pd.read_csv(self.trajectory_file)
        self.traj_df = self.traj_df.sort_values("tracking_timestamp_us")
        assert (
            self.traj_df["graph_uid"].nunique() == 1
        ), f"Seq({name})'s trajectory file contains multiple graphs."

        deltas_us = np.diff(self.traj_df["tracking_timestamp_us"])
        self.rate_hz = 1 / np.mean(deltas_us / 1000_000)

        # load in semi-dense point files if available, the results are stored as maps for quick querying by timestamp
        (self.uid_to_p3, self.uid_to_inv_dist_std) = self._load_semidense_global_points(
            semidense_global_point_file, compression_method=semidense_compression_method
        )
        (self.time_to_uids, self.uid_to_times) = self._load_semidense_observations(
            semidense_observation_file, compression_method=semidense_compression_method
        )

    def get_timestamps_ns(self):
        return self.traj_df["tracking_timestamp_us"].values * 1000

    def get_rate_hz(self):
        return self.rate_hz

    def get_nearest_poses(
        self,
        timestamps_ns: Union[np.ndarray, List[int]],
        tolerance_ns: int = 150_000,
        only_return_valid: bool = False,
    ):
        """
        Get the nearest poses to a list of timestamps.
        Args:
            timestamp_ns: A list of timestamps in nanoseconds.
            tolerance_ns: The tolerance for finding the nearest pose.
                If the difference between the input timestamp and the pose timestamp
                is less than this value, then it will be considered as the same pose.
        Returns:
            A list of translation and quaternion with the same length as `timestamps`.
                Translation from camera to world coordinate frame. 3x1
                Quaternion(xyzw) from camera to world coordinate frame. 4x1
                Gravity in world coordinate frame. 3x1
        """
        matched_df = self._find_matching_timestamps_in_df(
            self.traj_df,
            timestamps_ns,
            tolerance_ns,
            only_return_valid,
            timestamp_key_in_df="tracking_timestamp_us",
        )

        return matched_df[
            [
                "tx_world_device",
                "ty_world_device",
                "tz_world_device",
                "qx_world_device",
                "qy_world_device",
                "qz_world_device",
                "qw_world_device",
                "gravity_x_world",
                "gravity_y_world",
                "gravity_z_world",
            ]
        ]

    def get_nearest_semidense_points(
        self,
        timestamps_ns: Union[np.ndarray, List[int]],
        tolerance_ns: int = 150_000,
        only_return_valid: bool = False,
    ) -> pd.DataFrame:
        """
        Get points_in_world close to timestamps.
        Args:
            timestamp_ns: A list of timestamps in nanoseconds, Kx1
            tolerance_ns: The tolerance for finding the nearest frame.
                If the difference between the query timestamp and the semidense point timestamp
                is less than this value, then it will be considered as the same frame.
        Returns:
            a list of Tensors of Nx3 to represent observable world points,
            and a list of Tensors of Nx1 to represent the standard deviation of the semidense points.
        """
        # create data frame to re-use timestamp matching function
        time_to_uids_df = pd.DataFrame(
            list(self.time_to_uids.items()), columns=["tracking_timestamp_us", "uids"]
        )

        matched_time_to_uids_df = self._find_matching_timestamps_in_df(
            time_to_uids_df,
            timestamps_ns,
            tolerance_ns,
            only_return_valid,
            timestamp_key_in_df="tracking_timestamp_us",
        )

        df = pd.DataFrame(columns=["points_world", "points_inv_dist_std"])
        # loop over all matched timestamps, and stack point_in_world into a Nx3 tensor
        for uid_list in matched_time_to_uids_df["uids"]:
            points_world_per_timestamp = []
            inv_dist_std_per_timestamp = []
            for uid in uid_list:
                if (uid in self.uid_to_p3) and (uid in self.uid_to_inv_dist_std):
                    points_world_per_timestamp.append(self.uid_to_p3[uid])
                    inv_dist_std_per_timestamp.append(self.uid_to_inv_dist_std[uid])
                else:
                    raise ValueError(
                        f"Point UID {uid} not found in global semidense point file!"
                    )
            # end for uid
            df = df.append(
                {
                    "points_world": torch.stack(points_world_per_timestamp),
                    "points_inv_dist_std": torch.tensor(inv_dist_std_per_timestamp),
                },
                ignore_index=True,
            )
        # end for uid_list

        return df

    def _find_matching_timestamps_in_df(
        self,
        data_frame: pd.DataFrame,
        timestamps_ns: Union[np.ndarray, List[int]],
        tolerance_ns: int,
        only_return_valid: bool = False,
        timestamp_key_in_df: str = "tracking_timestamp_us",
    ) -> pd.DataFrame:
        """
        Helper function that given a list of timestamps, find the rows containing matching timestamps in the data frame, with some tolerance.
        Returns:
            a data frame containing the matching rows.
        """
        if isinstance(timestamps_ns, List):
            timestamps_ns = np.array(timestamps_ns)

        timestamps_us_df = pd.DataFrame(
            {timestamp_key_in_df: pd.Series(np.round(timestamps_ns / 1000).astype(int))}
        )
        timestamps_us_df = timestamps_us_df.sort_values(timestamp_key_in_df)

        # Use merge_asof to find the nearest pose for each timestamp.
        merged_df = pd.merge_asof(
            timestamps_us_df,
            data_frame,
            on=timestamp_key_in_df,
            tolerance=int(round(tolerance_ns / 1000)),
            direction="nearest",
        )

        if not only_return_valid:
            return merged_df
        else:
            # Check percentage of invalid timestamps.
            valid_merged_df = merged_df.dropna()

            invalid_count = len(merged_df) - len(valid_merged_df)
            invalid_percent = (invalid_count / len(merged_df)) * 100
            if invalid_percent > 5:
                logger.warning(
                    f"{invalid_count} ({invalid_percent:.2f}%) of the input timestamps can not find corresponding poses."
                )

            return valid_merged_df

    def _load_semidense_global_points(
        self,
        path: str,
        compression_method: Optional[str],
    ):
        print(f"loading global semi-dense points from {path}")
        uid_to_p3 = {}
        uid_to_inv_dist_std = {}

        with open(path, "rb") as f:
            csv_data = pd.read_csv(f, compression=compression_method)

            # select points and uids and return mapping
            uid_pts = csv_data[
                ["uid", "inv_dist_std", "px_world", "py_world", "pz_world"]
            ]

            for row in uid_pts.values:
                uid = int(row[0])
                inv_dist_std = float(row[1])
                p3 = torch.from_numpy(row[2:]).float()
                uid_to_p3[uid] = p3
                uid_to_inv_dist_std[uid] = inv_dist_std

        return uid_to_p3, uid_to_inv_dist_std

    def _load_semidense_observations(
        self, path: str, compression_method: Optional[str] = None
    ):
        """
        Load semidense observations from a csv file, returns two-way mapping between timestamp_in_us and point uids.
        Args:
            path: The path to the csv file.
        Returns:
            A tuple of two dictionaries.
            The first dictionary maps from timestamp to a list of uids.
            The second dictionary maps from uid to a list of timestamps.
        """

        logger.info(f"loading semidense observations from {path}")
        time_to_uids = defaultdict(list)
        uid_to_times = defaultdict(list)

        with open(path, "rb") as f:
            csv = pd.read_csv(f, compression=compression_method)
            csv = csv[["uid", "frame_tracking_timestamp_us"]]
            for row in csv.values:
                uid = int(row[0])
                time_ns = int(row[1])
                time_to_uids[time_ns].append(uid)
                uid_to_times[uid].append(time_ns)
        return time_to_uids, uid_to_times



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argparse import Namespace
from typing import Dict

from atek_v1.dataset.omni3d_adapter import (
    create_omni3d_raw_dataset,
    create_omni3d_webdataset,
    ObjectDetectionMode,
)

from torch.utils.data import Dataset


def create_inference_dataset(
    data_path: str, args: Namespace, model_config: Dict
) -> Dataset:
    """
    Create dataset for inference pipeline.

    Args:
        data_path (str): path to a sequence
        args (Namespace): args with options related to dataset creation
        model_config (Dict): model config with options related to dataset creation

    Return:
        dataset (Dataset): a dataset instance for the inference pipeline
    """

    if args.model_name == "cubercnn":
        if args.data_type == "wds":
            dataset = create_omni3d_webdataset(
                [data_path],
                batch_size=1,
                repeat=False,
                category_id_remapping_json=model_config["cfg"].ID_MAP_JSON,
                object_detection_mode=ObjectDetectionMode[
                    model_config["cfg"].DATASETS.OBJECT_DETECTION_MODE
                ],
            )
        elif args.data_type == "raw":
            dataset = create_omni3d_raw_dataset(
                data_path,
                selected_device_number=0,
                rotate_image_cw90deg=args.rotate_image_cw90deg,
                target_image_resolution=(args.width, args.height),
            )
        else:
            raise ValueError(
                f"Unknown input data type for building dataset: {args.data_type}"
            )
    else:
        raise ValueError(f"Unknown model name for building dataset: {args.model_name}")

    return dataset



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Union

import numpy as np
import torch

from atek_v1.utils.transform_utils import (
    batch_inverse,
    batch_transform_points,
    get_cuboid_corners,
)
from scipy.spatial import ConvexHull


BBOX_LINES_3D = [
    [0, 1],
    [1, 2],
    [2, 3],
    [3, 0],
    [4, 5],
    [5, 6],
    [6, 7],
    [7, 4],
    [0, 4],
    [1, 5],
    [2, 6],
    [3, 7],
]

BBOX_FACES_3D = [
    [4, 5, 6, 7],  # top face zmax
    [0, 1, 2, 3],  # bottom face zmin
    [2, 3, 7, 6],  # left face ymax
    [0, 1, 5, 4],  # right face ymin
    [1, 2, 6, 5],  # front face xmax
    [3, 0, 4, 7],  # back face xmin
]


def is_point_on_line_segment(point, start_point, end_point):
    """
    Check if a 3D point lies on a line segment defined by two other 3D points.
    """
    # Calculate the vectors from the start point to the end point and from the start point to the test point
    vector1 = end_point - start_point
    vector2 = point - start_point

    # Check if the cross product of the two vectors is approximately zero
    cross_product = torch.cross(vector1, vector2)
    if torch.allclose(cross_product, torch.zeros(3, dtype=point.dtype)):
        # Check if the dot product of the two vectors is positive and less than the squared length of vector1
        dot_product = torch.dot(vector1, vector2)
        if 0 <= dot_product <= torch.dot(vector1, vector1):
            return True

    return False


def is_point_inside_triangle(p, a, b, c):
    """
    Check if a 3D point lies inside a triangle defined by three other 3D points.
    """
    # Compute vectors
    v0 = c - a
    v1 = b - a
    v2 = p - a

    # Compute dot products
    dot00 = torch.dot(v0, v0)
    dot01 = torch.dot(v0, v1)
    dot02 = torch.dot(v0, v2)
    dot11 = torch.dot(v1, v1)
    dot12 = torch.dot(v1, v2)

    # Compute barycentric coordinates
    inv_denom = 1 / (dot00 * dot11 - dot01 * dot01)
    u = (dot11 * dot02 - dot01 * dot12) * inv_denom
    v = (dot00 * dot12 - dot01 * dot02) * inv_denom

    # Check if the point is in the triangle
    return (u >= 0) and (v >= 0) and (u + v <= 1)


def is_point_inside_plane(intersection_point, plane_points):
    """
    Check if a point lies inside a plane defined by four points.
    """
    if (
        is_point_inside_triangle(
            intersection_point, plane_points[0], plane_points[1], plane_points[2]
        )
        or is_point_inside_triangle(
            intersection_point, plane_points[0], plane_points[1], plane_points[3]
        )
        or is_point_inside_triangle(
            intersection_point, plane_points[0], plane_points[2], plane_points[3]
        )
        or is_point_inside_triangle(
            intersection_point, plane_points[3], plane_points[1], plane_points[2]
        )
        or is_point_on_line_segment(
            intersection_point, plane_points[0], plane_points[1]
        )
        or is_point_on_line_segment(
            intersection_point, plane_points[1], plane_points[2]
        )
        or is_point_on_line_segment(
            intersection_point, plane_points[2], plane_points[3]
        )
        or is_point_on_line_segment(
            intersection_point, plane_points[3], plane_points[0]
        )
    ):
        return True
    return False


def line_plane_intersection(
    plane_points: torch.Tensor, line_points: torch.Tensor
) -> Union[None, torch.Tensor]:
    """
    Compute the intersection point of a line and a plane.

    Args:
        plane_points: (4, 3) tensor, four points on the plane
        line_points: (2, 3) tensor, two points on the line

    Returns:
        intersection_point: (3,) tensor, intersection point
    """
    assert plane_points.size() == (4, 3)
    assert line_points.size() == (2, 3)

    # Compute the normal of the plane
    normal = torch.cross(
        plane_points[1] - plane_points[0], plane_points[2] - plane_points[0]
    )
    D = -torch.dot(normal, plane_points[0])

    # Compute the intersection point of line and plane
    P1, P2 = line_points[0], line_points[1]
    t = -(torch.dot(normal, P1) + D) / (torch.dot(normal, P2 - P1) + 1e-9)
    if 0 <= t.item() <= 1:
        intersection_point = P1 + t * (P2 - P1)

        # Check if point is inside plane
        if is_point_inside_plane(intersection_point, plane_points):
            return intersection_point

    return None


def check_coplanar(points: np.ndarray) -> bool:
    # check if the points are coplanar
    rank = np.linalg.matrix_rank(points - np.expand_dims(points[0], 0))
    return rank < 3


def iou_giou_single(
    cuboid_A_dimensions: torch.Tensor,
    T_world_A: torch.Tensor,
    cuboid_B_dimensions: torch.Tensor,
    T_world_B: torch.Tensor,
) -> Tuple[float, float, float]:
    """
    Computes IOU and GIOU between two 3D bounding boxes.

    Args:
        cuboid_A_dimensions: (3,) tensor, dimension of cuboid A
        T_world_A: (3, 4) tensor, transformation matrix from object to world of cuboid A
        cuboid_B_dimensions: (3,) tensor, dimension of cuboid B
        T_world_B: (3, 4) tensor, transformation matrix from object to world of cuboid B

    Returns:
        vol_in: float, volume of intersection
        iou: float, intersection over union
        giou: float, generalized intersection over union
    """
    assert cuboid_A_dimensions.shape == (3,)
    assert T_world_A.shape == (3, 4)
    assert cuboid_B_dimensions.shape == (3,)
    assert T_world_B.shape == (3, 4)

    half_extents_A = cuboid_A_dimensions.unsqueeze(0) / 2
    half_extents_B = cuboid_B_dimensions.unsqueeze(0) / 2

    cuboid_A_corners_in_world = batch_transform_points(
        get_cuboid_corners(half_extents_A),
        T_world_A.unsqueeze(0),
    )
    cuboid_B_corners_in_world = batch_transform_points(
        get_cuboid_corners(half_extents_B),
        T_world_B.unsqueeze(0),
    )

    # compute convex hull for A and B
    convex_hull_A = ConvexHull(cuboid_A_corners_in_world.squeeze(0).numpy())
    convex_hull_B = ConvexHull(cuboid_B_corners_in_world.squeeze(0).numpy())

    # compute convex hull for all points from A and B
    all_points = np.row_stack(
        [
            cuboid_A_corners_in_world.squeeze(0).numpy(),
            cuboid_B_corners_in_world.squeeze(0).numpy(),
        ]
    )
    convex_hull_all = ConvexHull(all_points)

    # find all the intersection points of A and B
    points_for_intersection = []

    cuboid_A_corners_in_B = batch_transform_points(
        cuboid_A_corners_in_world,
        batch_inverse(T_world_B.unsqueeze(0)),
    )
    cuboid_B_corners_in_A = batch_transform_points(
        cuboid_B_corners_in_world,
        batch_inverse(T_world_A.unsqueeze(0)),
    )

    A_inside_B_idx = (
        (-half_extents_B < cuboid_A_corners_in_B)
        & (cuboid_A_corners_in_B < half_extents_B)
    ).all(dim=-1)
    cuboid_A_corners_inside_B_in_world = cuboid_A_corners_in_world[A_inside_B_idx]
    if len(cuboid_A_corners_inside_B_in_world) > 0:
        points_for_intersection.append(cuboid_A_corners_inside_B_in_world.numpy())

    B_inside_A_idx = (
        (-half_extents_A < cuboid_B_corners_in_A)
        & (cuboid_B_corners_in_A < half_extents_A)
    ).all(dim=-1)
    cuboid_B_corners_inside_A_in_world = cuboid_B_corners_in_world[B_inside_A_idx]
    if len(cuboid_B_corners_inside_A_in_world) > 0:
        points_for_intersection.append(cuboid_B_corners_inside_A_in_world.numpy())

    edge_intersection_A_in_world = []
    edge_intersection_B_in_world = []
    for LINE in BBOX_LINES_3D:
        for FACE in BBOX_FACES_3D:
            plane_points_A_in_world = cuboid_A_corners_in_world.squeeze()[FACE]
            plane_points_B_in_world = cuboid_B_corners_in_world.squeeze()[FACE]

            line_points_A_in_world = cuboid_A_corners_in_world.squeeze()[LINE]
            line_points_B_in_world = cuboid_B_corners_in_world.squeeze()[LINE]

            intersect_points_in_world_A = line_plane_intersection(
                plane_points_B_in_world, line_points_A_in_world
            )
            if intersect_points_in_world_A is not None:
                edge_intersection_A_in_world.append(intersect_points_in_world_A)

            intersect_points_in_world_B = line_plane_intersection(
                plane_points_A_in_world, line_points_B_in_world
            )
            if intersect_points_in_world_B is not None:
                edge_intersection_B_in_world.append(intersect_points_in_world_B)

    if len(edge_intersection_A_in_world) > 0:
        points_for_intersection.append(
            torch.stack(edge_intersection_A_in_world).numpy()
        )
    if len(edge_intersection_B_in_world) > 0:
        points_for_intersection.append(
            torch.stack(edge_intersection_B_in_world).numpy()
        )

    # compute convex hull for intersection of A and B
    # if there are no intersection points, then the volume is zero
    if len(points_for_intersection) > 0:
        all_intersection_points_in_world = np.row_stack(points_for_intersection)

        # if all points are coplanar, then the volume is zero
        if check_coplanar(np.unique(all_intersection_points_in_world, axis=0)):
            vol_in = 0
        else:
            ch_intersect = ConvexHull(all_intersection_points_in_world)
            vol_in = ch_intersect.volume
    else:
        vol_in = 0

    # compute volume of union of A and B, IoU, GIoU
    AUB = convex_hull_A.volume + convex_hull_B.volume - vol_in
    iou = vol_in / AUB
    giou = iou - (convex_hull_all.volume - AUB) / convex_hull_all.volume

    return vol_in, iou, giou


def iou_giou(
    cuboid_A_dimensions: torch.Tensor,
    T_world_A: torch.Tensor,
    cuboid_B_dimensions: torch.Tensor,
    T_world_B: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Computes pairwise IOU and GIOU between M and N 3D bounding boxes.

    Args:
        cuboid_A_dimensions: (M, 3) tensor, dimension of cuboids A
        T_world_A: (M, 3, 4) tensor, transformation matrices from object to world of cuboids A
        cuboid_B_dimensions: (N, 3) tensor, dimension of cuboids B
        T_world_B: (N, 3, 4) tensor, transformation matrices from object to world of cuboids B

    Returns:
        pw_vol_in: (M, N) tensor, pairwise volumes of intersections
        pw_iou: (M, N) tensor, pairwise IOUs
        pw_giou: (M, N) tensor, pairwise GIOUs
    """

    assert cuboid_A_dimensions.ndim == cuboid_B_dimensions.ndim == 2
    assert cuboid_A_dimensions.shape[1] == cuboid_B_dimensions.shape[1] == 3
    assert T_world_A.ndim == T_world_B.ndim == 3
    assert T_world_A.shape[1:] == T_world_B.shape[1:] == (3, 4)

    M, N = len(cuboid_A_dimensions), len(cuboid_B_dimensions)
    dim_A_repeat = cuboid_A_dimensions.unsqueeze(1).repeat(1, N, 1)
    dim_B_repeat = cuboid_B_dimensions.unsqueeze(0).repeat(M, 1, 1)
    T_world_A_repeat = T_world_A.unsqueeze(1).repeat(1, N, 1, 1)
    T_world_B_repeat = T_world_B.unsqueeze(0).repeat(M, 1, 1, 1)

    dim_A_repeat = dim_A_repeat.reshape((M * N, 3))
    dim_B_repeat = dim_B_repeat.reshape((M * N, 3))
    T_world_A_repeat = T_world_A_repeat.reshape((M * N, 3, 4))
    T_world_B_repeat = T_world_B_repeat.reshape((M * N, 3, 4))

    # TODO: add batched implementation of this function to accelerate
    pw_vol_in, pw_iou, pw_giou = [], [], []
    for dim_A, T_A, dim_B, T_B in zip(
        dim_A_repeat, T_world_A_repeat, dim_B_repeat, T_world_B_repeat
    ):
        vol_in, iou, giou = iou_giou_single(dim_A, T_A, dim_B, T_B)
        pw_vol_in.append(vol_in)
        pw_iou.append(iou)
        pw_giou.append(giou)
    pw_vol_in = torch.Tensor(pw_vol_in).reshape((M, N))
    pw_iou = torch.Tensor(pw_iou).reshape((M, N))
    pw_giou = torch.Tensor(pw_giou).reshape((M, N))

    return pw_vol_in, pw_iou, pw_giou


def diagonal_error(
    pred_scale: torch.Tensor, target_scale: torch.Tensor
) -> torch.Tensor:
    """
    Compute the diagonal error between pred_scale and target_scale.

    Args:
        pred_scale: Tensor of shape (M, 3) representing the predicted scale.
        target_scale: Tensor of shape (N, 3) representing the target scale.

    Returns:
        Tensor of shape (M, N) containing the pairwise diagonal error.
    """
    assert pred_scale.ndim == target_scale.ndim == 2
    assert pred_scale.shape[1:] == target_scale.shape[1:]

    M, N = pred_scale.shape[0], target_scale.shape[0]

    # shape: (M, N, 3)
    pred_scale_repeat = pred_scale.unsqueeze(1).repeat((1, N, 1))
    target_scale_repeat = target_scale.unsqueeze(0).repeat((M, 1, 1))

    # shape: (M, N)
    pred_diag = torch.norm(pred_scale_repeat, p=2, dim=-1)
    target_diag = torch.norm(target_scale_repeat, p=2, dim=-1)

    diag_error = torch.abs(pred_diag - target_diag)

    return diag_error



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from dataclasses import asdict
from enum import Enum
from typing import Callable, Dict, List, Optional, Tuple

import numpy as np
import torch

from atek_v1.data_preprocess.data_schema import Frame
from atek_v1.dataset.atek_raw_dataset import AriaStreamIds, AtekRawFrameDataset
from atek_v1.dataset.atek_webdataset import create_atek_webdataset
from atek_v1.utils.camera_utils import linear_cam_matrix_from_intrinsics

from detectron2.structures import Boxes, Instances
from webdataset.filters import pipelinefilter
from webdataset.shardlists import single_node_only

KEY_MAPPING = {
    "f#214-1+image": "images",
    "F#214-1+camera_parameters": "camera_params",
    "F#214-1+Ts_camera_object": "Ts_camera_object",
    "F#214-1+object_dimensions": "object_dimensions",
    "F#214-1+bb2ds": "bb2ds_x0x1y0y1",
    "F#214-1+category_id_to_name": "category_id_to_name",
    "F#214-1+object_category_ids": "object_category_ids",
    "F#214-1+object_instance_ids": "object_instance_ids",
    "F#214-1+sequence_name": "sequence_name",
    "F#214-1+frame_id": "frame_id",
    "F#214-1+timestamp_ns": "timestamp_ns",
    "F#214-1+T_world_camera": "T_world_camera",
    "F#214-1+Ts_world_object": "Ts_world_object",
    "F#214-1+object_dimensions": "object_dimensions",
}


class ObjectDetectionMode(Enum):
    """
    An enum class to indicate if detection is per-category, or per instance:
    - PER_CATEGORY: will use the `category_id` (or `category_name`) field as the output label.
    - PER_INSTANCE: will use the `instance_id` field as the output label.
    """

    PER_CATEGORY = 1
    PER_INSTANCE = 2


def omni3d_key_selection(key: str) -> bool:
    return key in KEY_MAPPING.keys()


def omni3d_key_remap(key: str) -> str:
    if key in KEY_MAPPING.keys():
        return KEY_MAPPING[key]
    else:
        return key


def get_id_map(id_map_json):
    with open(id_map_json, "r") as f:
        id_map = json.load(f)
        id_map = {k: int(v) for k, v in id_map.items()}
    return id_map


def atek_to_omni3d(
    data,
    category_id_remapping: Optional[Dict] = None,
    object_detection_mode: ObjectDetectionMode = ObjectDetectionMode.PER_CATEGORY,
    min_bb2d_area=100,
    min_bb3d_depth=0.3,
    max_bb3d_depth=5,
):
    """
    A helper data transform function to convert the ATEK webdataset data to omni3d unbatched
    samples. Yield one unbatched sample a time to use the collation and batching mechanism in
    the webdataset properly.
    """
    for sample in data:
        # Image rgb->bgr->255->uint8
        assert sample["images"].shape[1] == 3 and sample["images"].ndim == 4
        images = (sample["images"].flip(1) * 255).to(torch.uint8)

        # Image information
        Ks = linear_cam_matrix_from_intrinsics(sample["camera_params"])
        image_height, image_width = sample["images"].shape[2:]

        for idx in range(len(images)):
            sample_new = {
                "image": images[idx],
                "K": Ks[idx].tolist(),  # CubeRCNN requires list input
                "height": image_height,
                "width": image_width,
                # Metadata
                "frame_id": sample["frame_id"][idx],
                "timestamp_ns": sample["timestamp_ns"][idx],
                "sequence_name": os.path.basename(
                    os.path.dirname(sample["sequence_name"][idx])
                ),
            }

            # Populate instances for omni3d
            instances = Instances(images.shape[2:])
            if sample["bb2ds_x0x1y0y1"][idx] is not None:
                bb2ds_x0y0x1y1 = sample["bb2ds_x0x1y0y1"][idx][:, [0, 2, 1, 3]]

                bb2ds_area = (bb2ds_x0y0x1y1[:, 2] - bb2ds_x0y0x1y1[:, 0]) * (
                    bb2ds_x0y0x1y1[:, 3] - bb2ds_x0y0x1y1[:, 1]
                )
                filter_bb2ds_area = bb2ds_area > min_bb2d_area

                bb3ds_depth = sample["Ts_camera_object"][idx][:, -1, -1]
                filter_bb3ds_depth = (min_bb3d_depth <= bb3ds_depth) & (
                    bb3ds_depth <= max_bb3d_depth
                )

                # Remap the semantic id and filter classes
                if object_detection_mode == ObjectDetectionMode.PER_CATEGORY:
                    # Use category names as category ids for category-based detection
                    sem_ids = [
                        sample["category_id_to_name"][0][str(cat_id)]
                        for cat_id in sample["object_category_ids"][0]
                    ]
                elif object_detection_mode == ObjectDetectionMode.PER_INSTANCE:
                    # Use instance ids as category ids for instance-based detection
                    sem_ids = [
                        str(inst_id) for inst_id in sample["object_instance_ids"][idx]
                    ]
                else:
                    raise ValueError(
                        f"Unsupported object detection mode: {object_detection_mode}"
                    )

                # Remap to Omni3D semantic ids, which should be from 0 to N.
                if category_id_remapping is not None:
                    sem_ids = [category_id_remapping.get(id, -1) for id in sem_ids]
                sem_ids = torch.tensor(sem_ids)
                filter_ids = sem_ids >= 0

                final_filter = filter_bb2ds_area & filter_bb3ds_depth & filter_ids
                Ts_camera_object_filtered = sample["Ts_camera_object"][idx][
                    final_filter
                ]
                ts_camera_object_filtered = Ts_camera_object_filtered[:, :, 3]

                filtered_projection_2d = (
                    Ks[idx].repeat(len(ts_camera_object_filtered), 1, 1)
                    @ ts_camera_object_filtered.unsqueeze(-1)
                ).squeeze(-1)
                filtered_projection_2d = filtered_projection_2d[
                    :, :2
                ] / filtered_projection_2d[:, 2].unsqueeze(-1)

                instances.gt_classes = sem_ids[final_filter]
                instances.gt_boxes = Boxes(bb2ds_x0y0x1y1[final_filter])
                instances.gt_poses = Ts_camera_object_filtered[:, :, :3]
                instances.gt_boxes3D = torch.cat(
                    [
                        filtered_projection_2d,
                        bb3ds_depth[final_filter].unsqueeze(-1),
                        # Omni3d has the inverted zyx dimensions
                        # https://github.com/facebookresearch/omni3d/blob/main/cubercnn/util/math_util.py#L144C1-L181C40
                        sample["object_dimensions"][idx][final_filter].flip(-1),
                        ts_camera_object_filtered,
                    ],
                    axis=-1,
                )

                sample_new["instances"] = instances
                filtered_category_ids = [
                    str(cat_id)
                    for cat_id, flag in zip(
                        sample["object_category_ids"][idx], final_filter
                    )
                    if flag
                ]
                category = [
                    sample["category_id_to_name"][idx][cat_id]
                    for cat_id in filtered_category_ids
                ]
                sample_new.update(
                    {
                        "T_world_camera": sample["T_world_camera"][idx],
                        "Ts_world_object": sample["Ts_world_object"][idx][final_filter],
                        "object_dimensions": sample["object_dimensions"][idx][
                            final_filter
                        ],
                        "bb2ds_x0x1y0y1": sample["bb2ds_x0x1y0y1"][idx][final_filter],
                        "category": category,
                    }
                )
                yield sample_new


def collate_as_list(batch):
    return list(batch)


def create_omni3d_webdataset(
    urls: List,
    batch_size: int,
    repeat: bool = False,
    nodesplitter: Callable = single_node_only,
    category_id_remapping_json: Optional[str] = None,
    object_detection_mode: ObjectDetectionMode = ObjectDetectionMode.PER_CATEGORY,
    min_bb2d_area=100,
    min_bb3d_depth=0.3,
    max_bb3d_depth=5,
):
    if category_id_remapping_json is not None:
        category_id_remapping = get_id_map(category_id_remapping_json)
    else:
        category_id_remapping = None

    return create_atek_webdataset(
        urls,
        batch_size,
        nodesplitter=nodesplitter,
        select_key_fn=omni3d_key_selection,
        remap_key_fn=omni3d_key_remap,
        data_transform_fn=pipelinefilter(atek_to_omni3d)(
            category_id_remapping,
            object_detection_mode,
            min_bb2d_area,
            min_bb3d_depth,
            max_bb3d_depth,
        ),
        collation_fn=collate_as_list,
        repeat=repeat,
    )


def atek_raw_to_omni3d(rgb_image_frame: Frame) -> List[Dict]:
    K = linear_cam_matrix_from_intrinsics(
        rgb_image_frame.camera_parameters[np.newaxis, ...]
    )[0]
    image = rgb_image_frame.image
    # image color RGB to BGR
    image = image[:, :, [2, 1, 0]]

    # image dimension (height, width, channel) to (channel, height, width)
    image = image.transpose(2, 0, 1)

    frame = {
        "data_source": rgb_image_frame.data_source,
        "sequence_name": rgb_image_frame.sequence_name.split("/")[-3],
        "frame_id": rgb_image_frame.frame_id,
        "timestamp_ns": rgb_image_frame.timestamp_ns,
        "T_world_camera": rgb_image_frame.T_world_camera,
        "image": torch.as_tensor(np.ascontiguousarray(image)),
        "height": image.shape[1],
        "width": image.shape[2],
        "K": K,
    }

    # GT object annotation info
    if "Ts_world_object" in asdict(rgb_image_frame).keys():
        frame["Ts_world_object"] = rgb_image_frame.Ts_world_object
        frame["object_dimensions"] = rgb_image_frame.object_dimensions
        frame["category"] = [
            rgb_image_frame.category_id_to_name[cat_id]
            for cat_id in rgb_image_frame.object_category_ids
        ]
        frame["bb2ds_x0x1y0y1"] = rgb_image_frame.bb2ds

    return [frame]


def create_omni3d_raw_dataset(
    raw_data_path: str,
    selected_device_number: int = 0,
    rotate_image_cw90deg: bool = True,
    target_image_resolution: Tuple[int, int] = (512, 512),
):
    atek_omni3d_raw_dataset = AtekRawFrameDataset(
        raw_data_path,
        selected_device_number=selected_device_number,
        stream_id=AriaStreamIds.rgb_stream_id,
        rotate_image_cw90deg=rotate_image_cw90deg,
        target_image_resolution=target_image_resolution,
        transform_fn=atek_raw_to_omni3d,
    )
    return atek_omni3d_raw_dataset



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Union

import torch

from atek_v1.utils.transform_utils import batch_transform_points, get_cuboid_corners
from pytorch3d.transforms import euler_angles_to_matrix


class Obb3:
    """
    A wrapper class for N 3D oriented bounding boxes (Obb3)
    """

    # size: (N, 3) tensor, object size in xyz order
    size: torch.Tensor
    # t_ref_obj: (N, 3) tensor, translation from object (obj) to reference (ref) coordinate in xyz order
    t_ref_obj: torch.Tensor
    # R_ref_obj: (N, 3, 3) tensor, rotation from object (obj) to reference (ref) coordinate
    R_ref_obj: torch.Tensor
    # instance id: (N,) tensor, unique instance id
    instance_id: torch.Tensor
    # category_id: (N,) tensor, category_id id
    category_id: torch.Tensor
    # score: (N,) tensor, confidence score
    score: torch.Tensor

    def __init__(
        self,
        size: torch.Tensor,
        t_ref_obj: torch.Tensor,
        R_ref_obj: torch.Tensor,
        instance_id: torch.Tensor,
        category_id: torch.Tensor,
        score: Optional[torch.Tensor] = None,
    ):
        assert (
            len(size)
            == len(t_ref_obj)
            == len(R_ref_obj)
            == len(instance_id)
            == len(category_id)
        )
        assert size.ndim == 2 and size.shape[1] == 3
        assert t_ref_obj.ndim == 2 and t_ref_obj.shape[1] == 3
        assert R_ref_obj.ndim == 3 and R_ref_obj.shape[1:] == (3, 3)
        assert instance_id.ndim == 1
        assert category_id.ndim == 1
        assert score is None or (score.ndim == 1 and len(score) == len(size))

        self.size = size
        self.t_ref_obj = t_ref_obj
        self.R_ref_obj = R_ref_obj
        self.instance_id = instance_id
        self.category_id = category_id
        if score is None:
            score = torch.ones((len(self.size),))
        self.score = score

    @property
    def T_ref_obj(self):
        return torch.cat((self.R_ref_obj, self.t_ref_obj.unsqueeze(-1)), dim=-1)

    # Nx8x3 corners in reference coordinate frame, reference can be world or camera
    @property
    def bb3_in_ref_frame(self):
        return batch_transform_points(get_cuboid_corners(self.size / 2), self.T_ref_obj)


def init_obb3(
    size: Union[List, torch.Tensor],
    t_ref_obj: Union[List, torch.Tensor],
    euler_angle: Union[List, torch.Tensor],
    instance_id: Union[List, torch.Tensor],
    category_id: Union[List, torch.Tensor],
    score: Optional[Union[List, torch.Tensor]] = None,
    convention: str = "XYZ",
) -> Obb3:
    """
    Initialize Obb3 instance with size, t_ref_obj, and euler_angle
    """
    if not isinstance(size, torch.Tensor):
        size = torch.Tensor(size)
    if not isinstance(t_ref_obj, torch.Tensor):
        t_ref_obj = torch.Tensor(t_ref_obj)
    if not isinstance(euler_angle, torch.Tensor):
        euler_angle = torch.Tensor(euler_angle)
    if not isinstance(instance_id, torch.Tensor):
        instance_id = torch.tensor(instance_id)
    if not isinstance(category_id, torch.Tensor):
        category_id = torch.tensor(category_id)
    if score is not None and not isinstance(score, torch.Tensor):
        score = torch.Tensor(score)
    R_ref_obj = euler_angles_to_matrix(euler_angle, convention=convention)
    return Obb3(
        size,
        t_ref_obj,
        R_ref_obj,
        instance_id,
        category_id,
        score=score,
    )
